from textwrap import dedent

import dill
import pytest

from dvcx.catalog import QUERY_DATASET_PREFIX
from dvcx.cli import query
from dvcx.error import QueryScriptRunError
from tests.utils import assert_row_names


@pytest.fixture
def catalog_info_filepath(cloud_test_catalog_tmpfile, tmp_path):
    catalog = cloud_test_catalog_tmpfile.catalog

    catalog_info = {
        "catalog_init_params": catalog.get_init_params(),
        "id_generator_params": catalog.id_generator.clone_params(),
        "metastore_params": catalog.metastore.clone_params(),
        "warehouse_params": catalog.warehouse.clone_params(),
    }
    catalog_info_filepath = tmp_path / "catalog-info"
    with open(catalog_info_filepath, "wb") as f:
        dill.dump(catalog_info, f)

    return catalog_info_filepath


def setup_catalog(query: str, catalog_info_filepath: str) -> str:
    query_catalog_setup = f"""\
    import dill
    from dvcx.catalog import Catalog

    catalog_info_filepath = {str(catalog_info_filepath)!r}
    with open(catalog_info_filepath, "rb") as f:
        catalog_info = dill.load(f)
    (
        id_generator_class,
        id_generator_args,
        id_generator_kwargs,
    ) = catalog_info["id_generator_params"]
    id_generator = id_generator_class(*id_generator_args, **id_generator_kwargs)
    (
        metastore_class,
        metastore_args,
        metastore_kwargs,
    ) = catalog_info["metastore_params"]
    metastore = metastore_class(*metastore_args, **metastore_kwargs)
    (
        warehouse_class,
        warehouse_args,
        warehouse_kwargs,
    ) = catalog_info["warehouse_params"]
    warehouse = warehouse_class(*warehouse_args, **warehouse_kwargs)
    catalog = Catalog(
        id_generator=id_generator,
        metastore=metastore,
        warehouse=warehouse,
        **catalog_info["catalog_init_params"],
    )
    """
    return dedent(query_catalog_setup + "\n" + query)


@pytest.mark.parametrize("cloud_type,version_aware", [("file", False)], indirect=True)
def test_query_cli(cloud_test_catalog_tmpfile, tmp_path, catalog_info_filepath, capsys):
    catalog = cloud_test_catalog_tmpfile.catalog
    src_uri = cloud_test_catalog_tmpfile.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery

    DatasetQuery({src_uri!r}, catalog=catalog)
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    filepath = tmp_path / "query_script.py"
    filepath.write_text(query_script)

    ds_name = "my-dataset"
    query(catalog, str(filepath), ds_name, columns=["name"])
    captured = capsys.readouterr()

    header, *rows = captured.out.splitlines()
    assert header.strip() == "name"
    name_rows = {row.split()[1] for row in rows}
    assert name_rows == {"cat1", "cat2", "description", "dog1", "dog2", "dog3", "dog4"}

    assert catalog.get_dataset(ds_name)


def test_query_cli_no_dataset_returned(
    cloud_test_catalog_tmpfile, tmp_path, catalog_info_filepath, capsys
):
    catalog = cloud_test_catalog_tmpfile.catalog

    query_script = """\
    from dvcx.query import DatasetQuery

    DatasetQuery("test", catalog=catalog)

    print("test")
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    filepath = tmp_path / "query_script.py"
    filepath.write_text(query_script)

    with pytest.raises(
        QueryScriptRunError,
        match="Last line in a script was not an instance of DatasetQuery",
    ):
        query(catalog, str(filepath), "my-dataset", columns=["name"])


@pytest.mark.parametrize(
    "save,save_as",
    (
        (True, None),
        (None, "my-dataset"),
        (True, "my-dataset"),
    ),
)
@pytest.mark.parametrize("save_dataset", (None, "new-dataset"))
def test_query(
    save,
    save_as,
    save_dataset,
    cloud_test_catalog_tmpfile,
    tmp_path,
    catalog_info_filepath,
):
    catalog = cloud_test_catalog_tmpfile.catalog
    src_uri = cloud_test_catalog_tmpfile.src_uri

    query_script = f"""\
    from dvcx.query import DatasetQuery
    ds = DatasetQuery({src_uri!r}, catalog=catalog)
    if {save_dataset!r}:
        ds = ds.save({save_dataset!r})
    ds
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    dataset, version, output, _ = catalog.query(
        query_script, save=save, save_as=save_as
    )
    if save_as:
        assert dataset.name == save_as
        assert catalog.get_dataset(save_as)
    elif save_dataset:
        assert dataset.name == save_dataset
        assert catalog.get_dataset(save_dataset)
    else:
        assert dataset.name.startswith(QUERY_DATASET_PREFIX)
    assert version == 1
    assert dataset.versions_values == [1]
    assert dataset.query_script == query_script
    assert_row_names(
        catalog,
        dataset,
        version,
        {
            "cat1",
            "cat2",
            "description",
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )


@pytest.mark.parametrize(
    "params,count",
    (
        (None, 7),
        ({"limit": 1}, 1),
        ({"limit": 5}, 5),
    ),
)
def test_query_params(
    params, count, cloud_test_catalog_tmpfile, tmp_path, catalog_info_filepath
):
    catalog = cloud_test_catalog_tmpfile.catalog
    src_uri = cloud_test_catalog_tmpfile.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery, param

    ds = DatasetQuery({src_uri!r}, catalog=catalog)
    if param("limit"):
        ds = ds.limit(int(param("limit")))
    ds
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    dataset, version, output, _ = catalog.query(query_script, save=True, params=params)
    assert len(list(catalog.ls_dataset_rows(dataset.name, version))) == count


def test_query_where_last_command_is_call_on_save_which_returns_attached_dataset(
    cloud_test_catalog_tmpfile, tmp_path, catalog_info_filepath
):
    """
    Testing use case where last command is call on DatasetQuery save which returns
    attached instance to underlying saved dataset
    """
    catalog = cloud_test_catalog_tmpfile.catalog
    src_uri = cloud_test_catalog_tmpfile.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery

    DatasetQuery({src_uri!r}, catalog=catalog).filter(C.name.glob("dog*")).save("dogs")
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    dataset, version, output, _ = catalog.query(query_script, save=True)
    assert not dataset.name.startswith(QUERY_DATASET_PREFIX)
    assert dataset.query_script == query_script
    assert version == 1
    assert dataset.versions_values == [1]
    assert_row_names(
        catalog,
        dataset,
        version,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )


def test_query_where_last_command_is_attached_dataset_query_created_from_save(
    cloud_test_catalog_tmpfile, tmp_path, catalog_info_filepath
):
    """
    Testing use case where last command is instance of DatasetQuery which is
    attached to underlying dataset by calling save just before
    """
    catalog = cloud_test_catalog_tmpfile.catalog
    src_uri = cloud_test_catalog_tmpfile.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery

    ds = DatasetQuery(
        {src_uri!r}, catalog=catalog
    ).filter(C.name.glob("dog*")).save("dogs")
    ds
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    dataset, version, output, _ = catalog.query(query_script, save=True)
    assert dataset.name == "dogs"
    assert dataset.query_script == query_script
    assert version == 1
    assert dataset.versions_values == [1]
    assert_row_names(
        catalog,
        dataset,
        version,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )


def test_query_where_last_command_is_attached_dataset_query_created_from_query(
    cloud_test_catalog_tmpfile, tmp_path, catalog_info_filepath
):
    """
    Testing use case where last command is instance of DatasetQuery which is
    attached to underlying dataset by creating query pointing to it
    """
    catalog = cloud_test_catalog_tmpfile.catalog
    src_uri = cloud_test_catalog_tmpfile.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery

    ds = DatasetQuery(
        {src_uri!r}, catalog=catalog
    ).filter(C.name.glob("dog*")).save("dogs")
    DatasetQuery(name="dogs", version=1, catalog=catalog)
    """
    query_script = setup_catalog(query_script, catalog_info_filepath)

    dataset, version, output, _ = catalog.query(query_script, save=True)
    assert dataset.name == "dogs"
    assert dataset.query_script == query_script
    assert version == 1
    assert dataset.versions_values == [1]
    assert_row_names(
        catalog,
        dataset,
        version,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )



from dvcx.query import C, DatasetQuery, udf
from dvcx.sql.types import Int


# Define the UDF:
@udf(
    ("name",),  # Columns consumed by the UDF.
    {
        "name_len": Int
    },  # Signals being returned by the UDF, with the signal name and type.
)
def name_len(name):
    if name.endswith(".json"):
        return (-1,)
    return (len(name),)


# Save as a new dataset.
DatasetQuery(
    path="gcs://dvcx-datalakes/dogs-and-cats/",
    anon=True,
).filter(C.name.glob("*cat*")).add_signals(name_len, parallel=-1).order_by(  # type: ignore[attr-defined]
    "name"
).limit(2).save("name_len")

# Output the contents of the new dataset.
print(DatasetQuery(name="name_len").select(C.name, C.name_len).results())



import hashlib
import tarfile
from functools import partial

from dvcx.sql.types import String

from .schema import C, DatasetRow, Object
from .udf import udf

md5 = partial(hashlib.md5, usedforsecurity=False)

__all__ = ["index_tar", "checksum"]


def load_tar(raw):
    with tarfile.open(fileobj=raw, mode="r:") as tar:
        return tar.getmembers()


@udf(
    (
        C.source,
        C.name,
        C.parent,
        C.size,
        C.vtype,
        C.dir_type,
        C.owner_name,
        C.owner_id,
        C.is_latest,
        C.last_modified,
        C.version,
        C.etag,
        Object(load_tar),
    ),
    DatasetRow.schema,
)
def index_tar(
    source,
    name,
    parent,
    size,
    vtype,
    dir_type,
    owner_name,
    owner_id,
    is_latest,
    last_modified,
    version,
    etag,
    tar_entries,
):
    # generate original tar files as well, along with subobjects
    yield DatasetRow.create(
        name,
        source=source,
        parent=parent,
        size=size,
        vtype=vtype,
        dir_type=dir_type,
        owner_name=owner_name,
        owner_id=owner_id,
        is_latest=bool(is_latest),
        last_modified=last_modified,
        version=version,
        etag=etag,
    )

    parent_path = name if not parent else f"{parent}/{name}"
    for info in tar_entries:
        if info.isfile():
            full_path = f"{parent_path}/{info.name}"
            parent_dir, name = full_path.rsplit("/", 1)
            yield DatasetRow.create(
                name,
                source=source,
                parent=parent_dir,
                size=info.size,
                vtype="tar",
                location={
                    "type": "tar",
                    "offset": info.offset_data,
                    "size": info.size,
                    "parent": parent_path,
                    "etag": etag,
                },
            )


BUFSIZE = 2**18


def file_digest(fileobj):
    """Calculate the digest of a file-like object."""
    buf = bytearray(BUFSIZE)  # Reusable buffer to reduce allocations.
    view = memoryview(buf)
    digestobj = md5()
    # From 3.11's hashlib.filedigest()
    while True:
        size = fileobj.readinto(buf)
        if size == 0:
            break  # EOF
        digestobj.update(view[:size])
    return digestobj.hexdigest()


@udf(params=[Object(file_digest)], output={"checksum": String})
def checksum(digest):
    return (digest,)



from textwrap import dedent
from typing import TYPE_CHECKING

from dvcx.catalog import Catalog

if TYPE_CHECKING:
    from dvcx.data_storage import AbstractWarehouse


def test_compile_query_script(catalog):
    script = dedent(
        """
        from dvcx.query import C, DatasetQuery, asUDF
        DatasetQuery("s3://bkt/dir1")
        """
    ).strip()
    result = catalog.compile_query_script(script)
    expected = dedent(
        """
        from dvcx.query import C, DatasetQuery, asUDF
        import dvcx.query.dataset
        dvcx.query.dataset.query_wrapper(
        DatasetQuery('s3://bkt/dir1'))
        """
    ).strip()
    assert result == expected


def test_catalog_warehouse_ready_callback(mocker, warehouse, id_generator, metastore):
    spy = mocker.spy(warehouse, "is_ready")

    def callback(warehouse: "AbstractWarehouse"):
        assert warehouse.is_ready()

    catalog = Catalog(
        id_generator, metastore, warehouse, warehouse_ready_callback=callback
    )

    spy.assert_not_called()

    _ = catalog.warehouse

    spy.assert_called_once()
    spy.reset_mock()

    _ = catalog.warehouse

    spy.assert_not_called()



from collections.abc import Sequence
from typing import Any, Union, get_args, get_origin

from pydantic import BaseModel, create_model

from dvcx.lib.feature import TYPE_TO_DVCX, Feature, ShallowFeature
from dvcx.lib.reader import FeatureReader
from dvcx.query.schema import Column
from dvcx.sql.types import NullType

FeatureLike = Union[type["Feature"], FeatureReader, Column, str]


class ColumnFeature(ShallowFeature):
    def _get_column_value(self):
        raise NotImplementedError("value is not defined for class ColumnFeature")


feature_registry: dict[type[BaseModel], type[Feature]] = {}


def pydantic_to_feature(data_cls: type[BaseModel]) -> type[Feature]:
    if data_cls in feature_registry:
        return feature_registry[data_cls]

    fields = {}
    for name, field_info in data_cls.model_fields.items():
        anno = field_info.annotation
        if anno not in TYPE_TO_DVCX:
            orig = get_origin(anno)
            if orig == list:
                anno = get_args(anno)  # type: ignore[assignment]
                if isinstance(anno, Sequence):
                    anno = anno[0]  # type: ignore[unreachable]
                is_list = True
            else:
                is_list = False

            try:
                Feature._convert_type(anno)
            except RuntimeError:
                if not issubclass(anno, Feature):  # type: ignore[arg-type]
                    anno = pydantic_to_feature(anno)  # type: ignore[arg-type]

            if is_list:
                anno = list[anno]  # type: ignore[valid-type]
        fields[name] = (anno, field_info.default)

    cls = create_model(
        data_cls.__name__,
        __base__=(data_cls, Feature),  # type: ignore[call-overload]
        **fields,
    )
    feature_registry[data_cls] = cls
    return cls


class FeatureTypes:
    @classmethod
    def column_class(
        cls,
        name: Union[str, Column],
        typ=Any,
        default=None,
        value_func=None,
    ):
        """Creating a column feature dynamically.
        :param fields:
            **name: <name> is string or a Column. For Column, a type can be specified.
            **typ: type of a column. Default is `Any`.
            **default: an optional default value
            **value_func: an optional function for get_value()
        """

        new_class = ColumnFeature

        if isinstance(name, Column):
            if typ is Any and not isinstance(name.type, NullType):
                typ = type(name.type)
            name = name.name

        fields = {name: (typ, default)}
        new_class_name = f"{new_class.__name__}_{name}"

        obj = create_model(
            new_class_name,
            __base__=new_class,  # type: ignore[call-overload]
            **fields,
        )

        obj._get_column_value = lambda self: getattr(self, name)

        if value_func:
            obj.get_value = lambda self: value_func(obj._get_column_value(self))
        else:
            obj.get_value = obj._get_column_value

        return obj

    @classmethod
    def column_classes(
        cls, fields: dict[Union[str, Column], tuple[type, Any]], value_func=None
    ) -> type:
        """Creating columns dynamically.
        :param fields:
            **fields: Attributes of the new model. They should be passed in the format:
            `<name>=(<type>, <default value>)` or `<name>=(<type>, <FieldInfo>)`
            where <name> is string or a Column
        """
        fields_text_keys = {
            key.name if isinstance(key, Column) else key: value
            for key, value in fields.items()
        }

        cls_suffix = "_".join(fields_text_keys.keys())
        new_class = ShallowFeature

        obj = create_model(
            f"{new_class.__name__}_{cls_suffix}",
            __base__=new_class,  # type: ignore[call-overload]
            **fields_text_keys,
        )

        obj._get_column_value = lambda self: tuple(
            [getattr(self, name) for name in fields_text_keys]
        )

        if value_func:
            obj.get_value = lambda self: value_func(self)
        else:
            obj.get_value = obj._get_column_value

        return obj

    @classmethod
    def _to_features(cls, *fr_classes: FeatureLike) -> Sequence[type["Feature"]]:
        features = []
        for fr in fr_classes:
            if isinstance(fr, (str, Column)):
                features.append(cls.column_class(fr))
            elif isinstance(fr, FeatureReader):
                features += cls._to_features(fr.fr_class)
            else:
                features.append(fr)
        return features



import pytest

from dvcx.node import DirType, Node
from dvcx.nodes_thread_pool import NodeChunk


@pytest.fixture
def nodes():
    return iter(
        [
            make_size_node(11, DirType.DIR, "a", "f1", 100),
            make_size_node(12, DirType.FILE, "b", "f2", 100),
            make_size_node(13, DirType.FILE, "c", "f3", 100),
            make_size_node(14, DirType.FILE, "d", "", 100),
            make_size_node(15, DirType.FILE, "e", "f5", 100),
            make_size_node(16, DirType.DIR, "f", "f6", 100),
            make_size_node(17, DirType.FILE, "g", "f7", 100),
        ]
    )


def make_size_node(node_id, dir_type, parent, name, size):
    return Node(
        node_id,
        "",  # vtype
        dir_type,
        parent,
        name,
        size=size,
    )


class FakeCache:
    def contains(self, _):
        return False


def make_chunks(nodes, *args, **kwargs):
    return NodeChunk(FakeCache(), "s3://foo", nodes, *args, **kwargs)


def test_node_bucket_the_only_item():
    bkt = make_chunks(iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 201)

    result = next(bkt)
    assert len(result) == 1
    assert next(bkt, None) is None


def test_node_bucket_the_only_item_over_limit():
    bkt = make_chunks(iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 1)

    result = next(bkt)
    assert len(result) == 1
    assert next(bkt, None) is None


def test_node_bucket_the_last_one():
    bkt = make_chunks(iter([make_size_node(20, DirType.FILE, 2, "file.csv", 100)]), 1)

    next(bkt)
    with pytest.raises(StopIteration):
        next(bkt)


def test_node_bucket_basic(nodes):
    bkt = list(make_chunks(nodes, 201))

    assert len(bkt) == 2
    assert len(bkt[0]) == 3
    assert len(bkt[1]) == 1


def test_node_bucket_full_split(nodes):
    bkt = list(make_chunks(nodes, 0))

    assert len(bkt) == 4
    assert len(bkt[0]) == 1
    assert len(bkt[1]) == 1
    assert len(bkt[2]) == 1
    assert len(bkt[3]) == 1



from typing import Any, cast

from botocore.exceptions import NoCredentialsError
from s3fs import S3FileSystem

from dvcx.node import Entry
from dvcx.utils import TIME_ZERO

from .fsspec import DELIMITER, Client

UPDATE_CHUNKSIZE = 1000


class ClientS3(Client):
    FS_CLASS = S3FileSystem
    PREFIX = "s3://"
    protocol = "s3"

    @classmethod
    def create_fs(cls, **kwargs) -> S3FileSystem:
        if "aws_endpoint_url" in kwargs:
            kwargs.setdefault("client_kwargs", {}).setdefault(
                "endpoint_url", kwargs.pop("aws_endpoint_url")
            )
        if "aws_key" in kwargs:
            kwargs.setdefault("key", kwargs.pop("aws_key"))
        if "aws_secret" in kwargs:
            kwargs.setdefault("secret", kwargs.pop("aws_secret"))
        if "aws_token" in kwargs:
            kwargs.setdefault("token", kwargs.pop("aws_token"))

        # caching bucket regions to use the right one in signed urls, otherwise
        # it tries to randomly guess and creates wrong signature
        kwargs.setdefault("cache_regions", True)

        # We want to use newer v4 signature version since regions added after
        # 2014 are not going to support v2 which is the older one.
        # All regions support v4.
        kwargs.setdefault("config_kwargs", {}).setdefault("signature_version", "s3v4")

        if not kwargs.get("anon"):
            try:
                # Run an inexpensive check to see if credentials are available
                super().create_fs(**kwargs).sign("s3://bucket/object")
            except NoCredentialsError:
                kwargs["anon"] = True
            except NotImplementedError:
                pass

        return cast(S3FileSystem, super().create_fs(**kwargs))

    async def _fetch_dir(
        self,
        prefix,
        pbar,
        result_queue,
    ):
        if prefix:
            prefix = prefix.lstrip(DELIMITER) + DELIMITER
        files = []
        subdirs = set()
        found = False
        async for info in self.fs._iterdir(self.name, prefix=prefix, versions=True):
            full_path = info["name"]
            _, subprefix, _ = self.fs.split_path(full_path)
            if prefix.strip(DELIMITER) == subprefix.strip(DELIMITER):
                found = True
                continue
            if info["type"] == "directory":
                name = full_path.split(DELIMITER)[-1]
                await result_queue.put(
                    [Entry.from_dir(prefix.rstrip("/"), name, last_modified=TIME_ZERO)]
                )
                subdirs.add(subprefix)
            else:
                files.append(self.convert_info(info, prefix.rstrip("/")))
                pbar.update()
            found = True
        if not found:
            raise FileNotFoundError(f"Unable to resolve remote path: {prefix}")
        if files:
            await result_queue.put(files)
        pbar.update(len(subdirs))
        return subdirs

    @staticmethod
    def clean_s3_version(ver):
        return ver if ver != "null" else ""

    def convert_info(self, v: dict[str, Any], parent: str) -> Entry:
        return Entry.from_file(
            parent=parent,
            name=v.get("Key", "").split(DELIMITER)[-1],
            etag=v.get("ETag", "").strip('"'),
            version=ClientS3.clean_s3_version(v.get("VersionId", "")),
            is_latest=v.get("IsLatest", True),
            last_modified=v.get("LastModified", ""),
            size=v["size"],
            owner_name=v.get("Owner", {}).get("DisplayName", ""),
            owner_id=v.get("Owner", {}).get("ID", ""),
        )



import base64
import pickle

from sqlalchemy import select

from dvcx.data_storage.serializer import deserialize
from dvcx.data_storage.sqlite import SQLiteDatabaseEngine, SQLiteIDGenerator


def get_rows(id_generator):
    uris = id_generator.db.execute(
        select(id_generator._table.c.uri, id_generator._table.c.last_id).select_from(
            id_generator._table
        )
    ).fetchall()
    return set(uris)


def test_init(sqlite_db):
    id_generator = SQLiteIDGenerator(sqlite_db)
    assert id_generator.db == sqlite_db
    assert id_generator._table_prefix is None
    assert sqlite_db.has_table("id_generator")
    assert get_rows(id_generator) == set()

    id_generator.cleanup_for_tests()
    assert not sqlite_db.has_table("id_generator")


def test_init_empty():
    id_generator = SQLiteIDGenerator()
    assert id_generator._table_prefix is None
    assert id_generator.db
    assert id_generator.db.has_table("id_generator")
    assert get_rows(id_generator) == set()

    id_generator.cleanup_for_tests()
    assert not id_generator.db.has_table("id_generator")


def test_init_with_prefix(sqlite_db):
    id_generator = SQLiteIDGenerator(sqlite_db, table_prefix="foo")
    assert id_generator.db == sqlite_db
    assert id_generator._table_prefix == "foo"
    assert sqlite_db.has_table("foo_id_generator")
    assert not sqlite_db.has_table("id_generator")
    assert get_rows(id_generator) == set()

    id_generator.cleanup_for_tests()
    assert not sqlite_db.has_table("foo_id_generator")


def test_clone(id_generator):
    clone = id_generator.clone()
    assert clone._table_prefix == id_generator._table_prefix
    assert get_rows(clone) == get_rows(id_generator)

    id_generator.init_id("foo")
    clone.get_next_id("bar")
    assert get_rows(id_generator) == {("foo", 0), ("bar", 1)}
    assert get_rows(clone) == get_rows(id_generator)

    id_generator.cleanup_for_tests()
    assert not id_generator.db.has_table("id_generator")
    assert not clone.db.has_table("id_generator")


def test_clone_params(id_generator):
    func, args, kwargs = id_generator.clone_params()
    clone = func(*args, **kwargs)
    assert clone._table_prefix == id_generator._table_prefix
    assert get_rows(clone) == get_rows(id_generator)

    id_generator.init_id("foo")
    clone.get_next_id("bar")
    assert get_rows(id_generator) == {("foo", 0), ("bar", 1)}
    assert get_rows(clone) == get_rows(id_generator)

    clone.cleanup_for_tests()
    assert not id_generator.db.has_table("id_generator")
    assert not clone.db.has_table("id_generator")


def test_serialize():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")

    obj = SQLiteIDGenerator(db, table_prefix="prefix")
    assert obj.db == db
    assert obj._table_prefix == "prefix"

    # Test clone
    obj2 = obj.clone()
    assert isinstance(obj2, SQLiteIDGenerator)
    assert obj2.db.db_file == obj.db.db_file
    assert obj2._table_prefix == "prefix"
    assert obj2.clone_params() == obj.clone_params()

    # Test serialization
    serialized = obj.serialize()
    assert serialized
    serialized_pickled = base64.b64decode(serialized.encode())
    assert serialized_pickled
    (f, args, kwargs) = pickle.loads(serialized_pickled)  # noqa: S301
    assert f == SQLiteIDGenerator.init_after_clone
    assert args == []
    assert kwargs == {
        "db_clone_params": db.clone_params(),
        "table_prefix": "prefix",
    }

    # Test deserialization
    obj3 = deserialize(serialized)
    assert isinstance(obj3, SQLiteIDGenerator)
    assert obj3.db.db_file == db.db_file
    assert obj3._table_prefix == "prefix"


def test_init_id(id_generator):
    assert get_rows(id_generator) == set()

    id_generator.init_id("foo")
    assert get_rows(id_generator) == {("foo", 0)}

    assert id_generator.get_next_id("foo") == 1
    assert get_rows(id_generator) == {("foo", 1)}

    id_generator.init_id("foo")  # second call should not raise an exception
    assert get_rows(id_generator) == {("foo", 1)}

    assert id_generator.get_next_id("foo") == 2
    assert get_rows(id_generator) == {("foo", 2)}

    id_generator.init_id("bar")
    assert get_rows(id_generator) == {("foo", 2), ("bar", 0)}


def test_get_next_id(id_generator):
    assert get_rows(id_generator) == set()

    assert id_generator.get_next_id("foo") == 1
    assert get_rows(id_generator) == {("foo", 1)}

    assert id_generator.get_next_id("foo") == 2
    assert get_rows(id_generator) == {("foo", 2)}

    assert id_generator.get_next_id("bar") == 1
    assert get_rows(id_generator) == {("foo", 2), ("bar", 1)}


def test_get_next_ids(id_generator):
    assert get_rows(id_generator) == set()

    assert id_generator.get_next_ids("foo", 3) == range(1, 4)
    assert get_rows(id_generator) == {("foo", 3)}

    assert id_generator.get_next_ids("foo", 20) == range(4, 24)
    assert get_rows(id_generator) == {("foo", 23)}

    assert id_generator.get_next_ids("bar", 1000) == range(1, 1001)
    assert get_rows(id_generator) == {("foo", 23), ("bar", 1000)}


def test_delete_uri(id_generator):
    assert get_rows(id_generator) == set()

    assert id_generator.get_next_ids("foo", 3) == range(1, 4)
    assert get_rows(id_generator) == {("foo", 3)}

    assert id_generator.get_next_ids("bar", 1000) == range(1, 1001)
    assert get_rows(id_generator) == {("foo", 3), ("bar", 1000)}

    id_generator.delete_uri("foo")
    assert get_rows(id_generator) == {("bar", 1000)}

    id_generator.delete_uri("bar")
    assert get_rows(id_generator) == set()


def test_delete_uris(id_generator):
    assert get_rows(id_generator) == set()

    assert id_generator.get_next_ids("foo", 3) == range(1, 4)
    assert get_rows(id_generator) == {("foo", 3)}

    assert id_generator.get_next_ids("bar", 1000) == range(1, 1001)
    assert get_rows(id_generator) == {("foo", 3), ("bar", 1000)}

    id_generator.delete_uris(["foo", "bar"])
    assert get_rows(id_generator) == set()



from sqlalchemy.sql.functions import GenericFunction

from dvcx.sql.types import Int64
from dvcx.sql.utils import compiler_not_implemented


class rand(GenericFunction):  # noqa: N801
    type = Int64()
    inherit_cache = True


compiler_not_implemented(rand)



import base64
import io
import os

import requests

try:
    from PIL import Image, ImageOps, UnidentifiedImageError
except ImportError as exc:
    raise ImportError(
        "Missing dependency Pillow for computer vision:\n"
        "To install run:\n\n"
        "  pip install 'dvcx[cv]'\n"
    ) from exc

from dvcx.query import Object, udf
from dvcx.sql.types import String

DEFAULT_FIT_BOX = (500, 500)
DEFAULT_TOKENS = 300


def encode_image(raw):
    try:
        img = Image.open(raw)
    except UnidentifiedImageError:
        return None
    img.load()
    img = ImageOps.fit(img, DEFAULT_FIT_BOX)
    output = io.BytesIO()
    img.save(output, format="JPEG")
    hex_data = output.getvalue()
    return base64.b64encode(hex_data).decode("utf-8")


@udf(
    params=(Object(encode_image),),  # Columns consumed by the UDF.
    output={
        "description": String,
        "error": String,
    },  # Signals being returned by the UDF.
    method="image_description",
)
class DescribeImage:
    def __init__(
        self,
        prompt="What is in this image?",
        max_tokens=DEFAULT_TOKENS,
        key="",
        timeout=30,
    ):
        if not key:
            key = os.getenv("OPENAI_API_KEY", "")
            if not key:
                raise ValueError(
                    "No key found. Please pass key or set the OPENAI_API_KEY "
                    "environment variable."
                )
        self.prompt = prompt
        self.max_tokens = max_tokens
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {key}",
        }
        self.timeout = timeout

    def image_description(self, base64_image):
        if base64_image is None:
            return ("", "Unknown image format")

        payload = {
            "model": "gpt-4-vision-preview",
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": self.prompt},
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{base64_image}"
                            },
                        },
                    ],
                }
            ],
            "max_tokens": self.max_tokens,
        }

        response = requests.post(
            "https://api.openai.com/v1/chat/completions",
            headers=self.headers,
            json=payload,
            timeout=self.timeout,
        )
        json_response = response.json()

        if "error" in json_response:
            error = str(json_response["error"])
            openai_description = ""
        else:
            error = ""
            openai_description = json_response["choices"][0]["message"]["content"]

        return (openai_description, error)



import json
import sqlite3

import ujson
from sqlalchemy import types

from dvcx.sql.types import TypeConverter, TypeReadConverter

try:
    import numpy as np

    numpy_imported = True
except ImportError:
    numpy_imported = False


class Array(types.UserDefinedType):
    cache_ok = True

    def __init__(self, item_type):
        self.item_type = item_type

    @property
    def python_type(self):
        return list

    def get_col_spec(self, **kwargs):
        return "ARRAY"


def adapt_array(arr):
    return ujson.dumps(arr)


def convert_array(arr):
    return ujson.loads(arr)


def adapt_np_array(arr):
    def _json_serialize(obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return obj

    if np.issubdtype(arr.dtype, np.object_):
        return json.dumps(arr.tolist(), default=_json_serialize)
    return ujson.dumps(arr.tolist())


def adapt_np_generic(val):
    return val.tolist()


def register_type_converters():
    sqlite3.register_adapter(list, adapt_array)
    sqlite3.register_converter("ARRAY", convert_array)
    if numpy_imported:
        sqlite3.register_adapter(np.ndarray, adapt_np_array)
        sqlite3.register_adapter(np.int32, adapt_np_generic)
        sqlite3.register_adapter(np.int64, adapt_np_generic)
        sqlite3.register_adapter(np.float32, adapt_np_generic)
        sqlite3.register_adapter(np.float64, adapt_np_generic)


class SQLiteTypeConverter(TypeConverter):
    def array(self, item_type):
        return Array(item_type)


class SQLiteTypeReadConverter(TypeReadConverter):
    def array(self, value, item_type, dialect):
        if isinstance(value, str):
            value = ujson.loads(value)
        return super().array(value, item_type, dialect)



import os
from unittest.mock import patch

import pytest

from dvcx.catalog.loader import (
    get_catalog,
    get_distributed_class,
    get_id_generator,
    get_metastore,
    get_warehouse,
)
from dvcx.data_storage.sqlite import (
    SQLiteDatabaseEngine,
    SQLiteIDGenerator,
    SQLiteMetastore,
    SQLiteWarehouse,
)
from dvcx.storage import StorageURI


class DistributedClass:
    def __init__(self, **kwargs):
        self.kwargs = kwargs


def test_get_id_generator():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")

    id_generator = SQLiteIDGenerator(db, table_prefix="prefix")
    assert id_generator.db == db
    assert id_generator._table_prefix == "prefix"

    with patch.dict(os.environ, {"DVCX__ID_GENERATOR": id_generator.serialize()}):
        id_generator2 = get_id_generator()
        assert id_generator2
        assert isinstance(id_generator2, SQLiteIDGenerator)
        assert id_generator2._db.db_file == db.db_file
        assert id_generator2._table_prefix == "prefix"
        assert id_generator2.clone_params() == id_generator.clone_params()

    with patch.dict(os.environ, {"DVCX__ID_GENERATOR": db.serialize()}):
        with pytest.raises(RuntimeError, match="instance of AbstractIDGenerator"):
            get_id_generator()


def test_get_metastore():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")
    id_generator = SQLiteIDGenerator(db, table_prefix="prefix")
    uri = StorageURI("s3://bucket")
    partial_id = 37

    metastore = SQLiteMetastore(id_generator, uri, partial_id, db)
    assert metastore.id_generator == id_generator
    assert metastore.uri == uri
    assert metastore.partial_id == partial_id
    assert metastore.db == db

    with patch.dict(os.environ, {"DVCX__METASTORE": metastore.serialize()}):
        metastore2 = get_metastore(None)
        assert metastore2
        assert isinstance(metastore2, SQLiteMetastore)
        assert metastore2.id_generator._db.db_file == metastore.id_generator._db.db_file
        assert (
            metastore2.id_generator._table_prefix
            == metastore.id_generator._table_prefix
        )
        assert metastore2.uri == uri
        assert metastore2.partial_id == partial_id
        assert metastore2.db.db_file == db.db_file
        assert metastore2.clone_params() == metastore.clone_params()

    with patch.dict(os.environ, {"DVCX__METASTORE": db.serialize()}):
        with pytest.raises(RuntimeError, match="instance of AbstractMetastore"):
            get_metastore(None)


def test_get_warehouse():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")
    id_generator = SQLiteIDGenerator(db, table_prefix="prefix")

    warehouse = SQLiteWarehouse(id_generator, db)
    assert warehouse.id_generator == id_generator
    assert warehouse.db == db

    with patch.dict(os.environ, {"DVCX__WAREHOUSE": warehouse.serialize()}):
        warehouse2 = get_warehouse(None)
        assert warehouse2
        assert isinstance(warehouse2, SQLiteWarehouse)
        assert warehouse2.id_generator._db.db_file == warehouse.id_generator._db.db_file
        assert (
            warehouse2.id_generator._table_prefix
            == warehouse.id_generator._table_prefix
        )
        assert warehouse2.db.db_file == db.db_file
        assert warehouse2.clone_params() == warehouse.clone_params()

    with patch.dict(os.environ, {"DVCX__WAREHOUSE": db.serialize()}):
        with pytest.raises(RuntimeError, match="instance of AbstractWarehouse"):
            get_warehouse(None)


def test_get_distributed_class():
    distributed_args = {"foo": "bar", "baz": "37", "empty": ""}
    env = {
        "DVCX_DISTRIBUTED": "tests.unit.test_catalog_loader.DistributedClass",
        "DVCX_DISTRIBUTED_ARG_FOO": "bar",
        "DVCX_DISTRIBUTED_ARG_BAZ": "37",
        "DVCX_DISTRIBUTED_ARG_EMPTY": "",
    }

    with patch.dict(os.environ, env):
        distributed = get_distributed_class()
        assert distributed
        assert isinstance(distributed, DistributedClass)
        assert distributed.kwargs == distributed_args

    with patch.dict(os.environ, {"DVCX_DISTRIBUTED": ""}):
        with pytest.raises(
            RuntimeError, match="DVCX_DISTRIBUTED import path is required"
        ):
            get_distributed_class()

    with patch.dict(
        os.environ,
        {"DVCX_DISTRIBUTED": "tests.unit.test_catalog_loader.NonExistent"},
    ):
        with pytest.raises(AttributeError, match="has no attribute 'NonExistent'"):
            get_distributed_class()

    with patch.dict(os.environ, {"DVCX_DISTRIBUTED": "DistributionClass"}):
        with pytest.raises(RuntimeError, match="Invalid DVCX_DISTRIBUTED import path"):
            get_distributed_class()


def test_get_catalog():
    db = SQLiteDatabaseEngine.from_db_file(":memory:")
    id_generator = SQLiteIDGenerator(db, table_prefix="prefix")
    uri = StorageURI("s3://bucket")
    partial_id = 73
    metastore = SQLiteMetastore(id_generator, uri, partial_id, db)
    warehouse = SQLiteWarehouse(id_generator, db)
    env = {
        "DVCX__ID_GENERATOR": id_generator.serialize(),
        "DVCX__METASTORE": metastore.serialize(),
        "DVCX__WAREHOUSE": warehouse.serialize(),
    }

    with patch.dict(os.environ, env):
        catalog = get_catalog()
        assert catalog

        assert catalog.id_generator
        assert isinstance(catalog.id_generator, SQLiteIDGenerator)
        assert catalog.id_generator._db.db_file == db.db_file
        assert catalog.id_generator._table_prefix == "prefix"
        assert catalog.id_generator.clone_params() == id_generator.clone_params()

        assert catalog.metastore
        assert isinstance(catalog.metastore, SQLiteMetastore)
        assert (
            catalog.metastore.id_generator._db.db_file
            == metastore.id_generator._db.db_file
        )
        assert (
            catalog.metastore.id_generator._table_prefix
            == metastore.id_generator._table_prefix
        )
        assert catalog.metastore.uri == uri
        assert catalog.metastore.partial_id == partial_id
        assert catalog.metastore.db.db_file == db.db_file
        assert catalog.metastore.clone_params() == metastore.clone_params()

        assert catalog.warehouse
        assert isinstance(catalog.warehouse, SQLiteWarehouse)
        assert (
            catalog.warehouse.id_generator._db.db_file
            == warehouse.id_generator._db.db_file
        )
        assert (
            catalog.warehouse.id_generator._table_prefix
            == warehouse.id_generator._table_prefix
        )
        assert catalog.warehouse.db.db_file == db.db_file
        assert catalog.warehouse.clone_params() == warehouse.clone_params()



# pip install torch
import torch

from dvcx.lib.hf_image_to_text import LLaVAdescribe
from dvcx.query import C, DatasetQuery

model = "llava-hf/llava-1.5-7b-hf"

# HuggingFace supports the following base models:
#
# "llava-hf/llava-1.5-7b-hf"
# "llava-hf/llava-1.5-13b-hf"
# "llava-hf/bakLlava-v1-hf"
#
# https://huggingface.co/llava-hf

source = "gcs://dvcx-datalakes/dogs-and-cats/"

# device='mps' not supported
if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

if __name__ == "__main__":
    results = (
        DatasetQuery(
            source,
            anon=True,
        )
        .filter(C.name.glob("cat*.jpg"))
        .limit(2)
        .add_signals(
            LLaVAdescribe(
                device=device,
                model=model,
            ),
            parallel=False,
        )
        .select("source", "parent", "name", "description", "error")
        .results()
    )
    print(*results, sep="\n")



from dvcx.data_storage.id_generator import AbstractDBIDGenerator, AbstractIDGenerator
from dvcx.data_storage.metastore import (
    AbstractDBMetastore,
    AbstractMetastore,
)
from dvcx.data_storage.warehouse import AbstractWarehouse

__all__ = [
    "AbstractDBIDGenerator",
    "AbstractIDGenerator",
    "AbstractDBMetastore",
    "AbstractMetastore",
    "AbstractWarehouse",
]



import os
from textwrap import dedent

import pytest

from dvcx.utils import (
    determine_processes,
    dvcx_paths_join,
    import_object,
    retry_with_backoff,
    sizeof_fmt,
    sql_escape_like,
    suffix_to_number,
)

DVCX_TEST_PATHS = ["/file1", "file2", "/dir/file3", "dir/file4"]
DVCX_EX_ROOT = ["/file1", "/file2", "/dir/file3", "/dir/file4"]
DVCX_EX_SUBDIR = [
    "subdir/file1",
    "subdir/file2",
    "subdir/dir/file3",
    "subdir/dir/file4",
]
DVCX_EX_DOUBLE_SUBDIR = [
    "subdir/double/file1",
    "subdir/double/file2",
    "subdir/double/dir/file3",
    "subdir/double/dir/file4",
]


@pytest.mark.parametrize(
    "src,paths,expected",
    (
        ("", DVCX_TEST_PATHS, DVCX_EX_ROOT),
        ("/", DVCX_TEST_PATHS, DVCX_EX_ROOT),
        ("/*", DVCX_TEST_PATHS, DVCX_EX_ROOT),
        ("/file*", DVCX_TEST_PATHS, DVCX_EX_ROOT),
        ("subdir", DVCX_TEST_PATHS, DVCX_EX_SUBDIR),
        ("subdir/", DVCX_TEST_PATHS, DVCX_EX_SUBDIR),
        ("subdir/*", DVCX_TEST_PATHS, DVCX_EX_SUBDIR),
        ("subdir/file*", DVCX_TEST_PATHS, DVCX_EX_SUBDIR),
        ("subdir/double", DVCX_TEST_PATHS, DVCX_EX_DOUBLE_SUBDIR),
        ("subdir/double/", DVCX_TEST_PATHS, DVCX_EX_DOUBLE_SUBDIR),
        ("subdir/double/*", DVCX_TEST_PATHS, DVCX_EX_DOUBLE_SUBDIR),
        ("subdir/double/file*", DVCX_TEST_PATHS, DVCX_EX_DOUBLE_SUBDIR),
    ),
)
def test_dvcx_paths_join(src, paths, expected):
    assert list(dvcx_paths_join(src, paths)) == expected


@pytest.mark.parametrize(
    "num,suffix,si,expected",
    (
        (1, "", False, "   1"),
        (536, "", False, " 536"),
        (1000, "", False, "1000"),
        (1000, "", True, "1.0K"),
        (1000, " tests", False, "1000 tests"),
        (1000, " tests", True, "1.0K tests"),
        (100000, "", False, "97.7K"),
        (100000, "", True, "100.0K"),
        (1000000, "", True, "1.0M"),
        (1000000000, "", True, "1.0G"),
        (1000000000000, "", True, "1.0T"),
        (1000000000000000, "", True, "1.0P"),
        (1000000000000000000, "", True, "1.0E"),
        (1000000000000000000000, "", True, "1.0Z"),
        (1000000000000000000000000, "", True, "1.0Y"),
        (1000000000000000000000000000, "", True, "1.0R"),
        (1000000000000000000000000000000, "", True, "1.0Q"),
    ),
)
def test_sizeof_fmt(num, suffix, si, expected):
    assert sizeof_fmt(num, suffix, si) == expected


@pytest.mark.parametrize(
    "text,expected",
    (
        ("1", 1),
        ("50", 50),
        ("1K", 1024),
        ("1k", 1024),
        ("2M", 1024 * 1024 * 2),
    ),
)
def test_suffix_to_number(text, expected):
    assert suffix_to_number(text) == expected


@pytest.mark.parametrize(
    "text",
    (
        "",
        "Bogus",
        "50H",
    ),
)
def test_suffix_to_number_invalid(text):
    with pytest.raises(ValueError):
        suffix_to_number(text)


@pytest.mark.parametrize(
    "text,expected",
    (
        ("test like", "test like"),
        ("Can%t \\escape_this", "Can\\%t \\\\escape\\_this"),
    ),
)
def test_sql_escape_like(text, expected):
    assert sql_escape_like(text) == expected


def test_import_object(tmp_path):
    fname = tmp_path / "foo.py"
    code = """\
        def hello():
            return "Hello!"
    """
    fname.write_text(dedent(code))
    func = import_object(f"{fname}:hello")
    assert func() == "Hello!"


def test_import_object_relative(tmp_path, monkeypatch):
    fname = tmp_path / "foo.py"
    code = """\
        def hello():
            return "Hello!"
    """
    fname.write_text(dedent(code))
    monkeypatch.chdir(tmp_path)
    func = import_object("foo.py:hello")
    assert func() == "Hello!"


def test_retry_with_backoff():
    called = 0
    retries = 2

    @retry_with_backoff(retries=retries, backoff_sec=0.05)
    def func_with_exception():
        nonlocal called
        called += 1
        raise RuntimeError("Error")

    with pytest.raises(RuntimeError):
        func_with_exception()
    assert called == retries + 1

    called = 0  # resetting called

    @retry_with_backoff(retries=retries, backoff_sec=0.05)
    def func_ok():
        nonlocal called
        called += 1

    func_ok()
    assert called == 1


@pytest.mark.parametrize(
    "parallel,settings,expected",
    (
        (None, None, False),
        (None, "-1", True),
        (None, "0", False),
        (None, "5", 5),
        (-1, "5", True),
        (0, "5", False),
        (10, "5", 10),
    ),
)
def test_determine_processes(parallel, settings, expected):
    if settings is not None:
        os.environ["DVCX_SETTINGS_PARALLEL"] = settings
    assert determine_processes(parallel) == expected



try:
    import pandas as pd
except ImportError:
    pd = None


class DvcxError(Exception):
    def __init__(self, message):
        super().__init__(message)


def row_to_pandas(args, params):
    data = dict(zip([i.name for i in params], args))
    return pd.Series(data, name=data.get("name"))


def row_list_to_pandas(args, params):
    return pd.DataFrame(args, columns=[i.name for i in params])


def bin_to_array(data):
    return [
        int.from_bytes(data[i : i + 4], byteorder="big") for i in range(0, len(data), 4)
    ]


def array_to_bin(integers):
    return b"".join(int.to_bytes(i, length=4, byteorder="big") for i in integers)


def union_dicts(*dicts):
    """Union dictionaries.
    Equivalent to `d1 | d2 | d3` in Python3.9+ but works in older versions.
    """
    result = None
    for d in dicts:
        if not isinstance(d, dict):
            raise TypeError("All arguments must be dictionaries.")
        if not result:
            result = d.copy()
        else:
            result.update(d)
    return result



import asyncio
import functools
from collections import Counter
from contextlib import contextmanager

import pytest
from fsspec.asyn import sync
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from dvcx.asyn import AsyncMapper, OrderedMapper, get_loop


async def fake_io(i):
    # print(f"task {i}")
    await asyncio.sleep(i)
    return i


def join_all_tasks(loop, timeout=4):
    tasks = asyncio.all_tasks(loop)
    if tasks:
        coro = asyncio.wait(tasks, timeout=timeout)
        future = asyncio.run_coroutine_threadsafe(coro, loop=loop)
        try:
            # Time out if any tasks are still running
            future.result(timeout=timeout)
        finally:
            # Avoid annoying warning after a timeout
            for t in asyncio.all_tasks(loop):
                t.cancel()


@contextmanager
def mock_time(loop):
    from aiotools import VirtualClock

    clock = VirtualClock()
    cm = clock.patch_loop()

    async def patch():
        cm.__enter__()

    async def unpatch():
        cm.__exit__(None, None, None)

    asyncio.run_coroutine_threadsafe(patch(), loop).result()
    try:
        yield
    finally:
        asyncio.run_coroutine_threadsafe(unpatch(), loop).result()


@pytest.fixture
def loop():
    loop = get_loop()
    with mock_time(loop):
        try:
            yield loop
        finally:
            join_all_tasks(loop)


@pytest.mark.parametrize("create_mapper", [AsyncMapper, OrderedMapper])
def test_mapper_fsspec(create_mapper, loop):
    n_rows = 50

    async def process(row):
        await mapper.to_thread(functools.partial(sync, loop, fake_io, row))
        return row

    mapper = create_mapper(process, range(n_rows), workers=10, loop=loop)
    result = [sync(loop, fake_io, i + n_rows) for i in mapper.iterate(timeout=4)]
    if mapper.order_preserving:
        assert result == list(range(n_rows, 2 * n_rows))
    else:
        assert set(result) == set(range(n_rows, 2 * n_rows))


@pytest.mark.parametrize("create_mapper", [AsyncMapper, OrderedMapper])
def test_mapper_generator_shutdown(create_mapper, loop):
    """
    Check that throwing an exception into AsyncMapper.iterate() terminates it cleanly.
    Note that finalising a generator involves throwing StopIteration into it.
    """

    async def process(row):
        await mapper.to_thread(functools.partial(sync, loop, fake_io, row))
        return row

    class MyError(Exception):
        pass

    mapper = create_mapper(process, range(50), workers=10, loop=loop)
    iterator = mapper.iterate(timeout=4)
    next(iterator)
    with pytest.raises(MyError):
        iterator.throw(MyError)


@pytest.mark.parametrize("create_mapper", [AsyncMapper, OrderedMapper])
def test_mapper_exception_while_processing(create_mapper, loop):
    async def process(row):
        await mapper.to_thread(functools.partial(sync, loop, fake_io, row))
        if row == 12:
            raise RuntimeError
        return row

    mapper = create_mapper(process, range(50), workers=10, loop=loop)
    with pytest.raises(RuntimeError):
        list(mapper.iterate(timeout=4))


@pytest.mark.parametrize("create_mapper", [AsyncMapper, OrderedMapper])
@settings(deadline=None)
@given(
    inputs=st.lists(st.integers(min_value=0, max_value=100), max_size=20),
    workers=st.integers(min_value=1, max_value=5),
)
def test_mapper_hypothesis(inputs, workers, create_mapper):
    async def process(input):
        await asyncio.sleep(input)
        return input

    loop = get_loop()
    mapper = create_mapper(process, inputs, workers=workers, loop=loop)
    with mock_time(loop):
        try:
            result = list(mapper.iterate(timeout=4))
        finally:
            join_all_tasks(loop)
    if mapper.order_preserving:
        assert result == inputs
    else:
        assert Counter(result) == Counter(inputs)


@pytest.mark.parametrize("create_mapper", [AsyncMapper, OrderedMapper])
@settings(deadline=None)
@given(
    inputs=st.lists(
        st.tuples(st.booleans(), st.integers(min_value=0, max_value=100)), max_size=20
    ),
    workers=st.integers(min_value=1, max_value=5),
)
def test_mapper_exception_hypothesis(inputs, workers, create_mapper):
    assume(any(n[0] for n in inputs))

    async def process(input):
        raising, n = input
        await asyncio.sleep(n)
        if raising:
            raise RuntimeError
        return input

    loop = get_loop()
    mapper = create_mapper(process, inputs, workers=workers, loop=loop)
    with mock_time(loop):
        try:
            with pytest.raises(RuntimeError):
                list(mapper.iterate(timeout=4))
        finally:
            join_all_tasks(loop)



import io
import json
import os
from contextlib import suppress
from pathlib import Path
from textwrap import dedent
from urllib.parse import urlparse

import pytest
import yaml

from dvcx.catalog import parse_edvcx_file
from dvcx.cli import garbage_collect
from dvcx.error import (
    QueryScriptCompileError,
    QueryScriptDatasetNotFound,
    QueryScriptRunError,
    StorageNotFoundError,
)
from tests.data import ENTRIES
from tests.utils import (
    DEFAULT_TREE,
    TARRED_TREE,
    assert_row_names,
    create_tar_dataset,
    make_index,
    skip_if_not_sqlite,
    tree_from_path,
)


@pytest.fixture
def pre_created_ds_name():
    return "pre_created_dataset"


@pytest.fixture
def mock_os_pipe(mocker):
    r, w = os.pipe()
    mocker.patch("os.pipe", return_value=(r, w))

    try:
        yield (r, w)
    finally:
        with suppress(OSError):
            os.close(r)
        with suppress(OSError):
            os.close(w)


@pytest.fixture
def mock_popen(mocker):
    m = mocker.patch(
        "subprocess.Popen", returncode=0, stdout=io.StringIO(), stderr=io.StringIO()
    )
    m.return_value.__enter__.return_value = m
    # keep in sync with the returncode
    m.poll.side_effect = lambda: m.returncode if m.poll.call_count > 1 else None
    return m


@pytest.fixture
def mock_popen_dataset_created(
    mock_popen, cloud_test_catalog, mock_os_pipe, listed_bucket
):
    # create dataset which would be created in subprocess
    ds_name = cloud_test_catalog.catalog.generate_query_dataset_name()
    ds_version = 1
    cloud_test_catalog.catalog.create_dataset_from_sources(
        ds_name,
        [f"{cloud_test_catalog.src_uri}/dogs/*"],
        recursive=True,
    )

    _, w = mock_os_pipe
    with open(w, mode="w", closefd=False) as f:
        f.write(json.dumps({"dataset": (ds_name, ds_version)}))

    mock_popen.configure_mock(stdout=io.StringIO("user log 1\nuser log 2"))
    yield mock_popen


@pytest.fixture
def fake_index(catalog):
    src = "s3://whatever"
    make_index(catalog, src, ENTRIES)
    return src


def test_find(catalog, fake_index):
    src_uri = fake_index
    assert set(catalog.find([src_uri])) == {
        f"{src_uri}/{entry.full_path}" for entry in ENTRIES
    }

    with pytest.raises(FileNotFoundError):
        set(catalog.find([f"{src_uri}/does_not_exist"]))


def test_find_names_paths_size_type(catalog, fake_index):
    src_uri = fake_index

    assert set(catalog.find([src_uri], names=["*cat*"])) == {
        f"{src_uri}/cats/",
        f"{src_uri}/cats/cat1",
        f"{src_uri}/cats/cat2",
    }

    assert set(catalog.find([src_uri], names=["*cat*"], typ="dir")) == {
        f"{src_uri}/cats/",
    }

    assert len(list(catalog.find([src_uri], names=["*CAT*"]))) == 0

    assert set(catalog.find([src_uri], inames=["*CAT*"])) == {
        f"{src_uri}/cats/",
        f"{src_uri}/cats/cat1",
        f"{src_uri}/cats/cat2",
    }

    assert set(catalog.find([src_uri], paths=["*cats/cat*"])) == {
        f"{src_uri}/cats/cat1",
        f"{src_uri}/cats/cat2",
    }

    assert len(list(catalog.find([src_uri], paths=["*caTS/CaT**"]))) == 0

    assert set(catalog.find([src_uri], ipaths=["*caTS/CaT*"])) == {
        f"{src_uri}/cats/cat1",
        f"{src_uri}/cats/cat2",
    }

    assert set(catalog.find([src_uri], size="5", typ="f")) == {
        f"{src_uri}/description",
    }

    assert set(catalog.find([src_uri], size="-3", typ="f")) == {
        f"{src_uri}/dogs/dog2",
    }


def test_find_names_columns(cloud_test_catalog, cloud_type):
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    owner = "webfile" if cloud_type == "s3" else ""

    assert set(
        catalog.find(
            [src_uri],
            names=["*cat*"],
            columns=["du", "name", "owner", "path", "size", "type"],
        )
    ) == {
        "\t".join(columns)
        for columns in [
            ["8", "cats", "", f"{src_uri}/cats/", "0", "d"],
            ["4", "cat1", owner, f"{src_uri}/cats/cat1", "4", "f"],
            ["4", "cat2", owner, f"{src_uri}/cats/cat2", "4", "f"],
        ]
    }


@pytest.mark.parametrize(
    "recursive,star,dir_exists",
    (
        (True, True, False),
        (True, False, False),
        (True, False, True),
        (False, True, False),
        (False, False, False),
    ),
)
def test_cp_root(cloud_test_catalog, recursive, star, dir_exists, cloud_type):
    src_uri = cloud_test_catalog.src_uri
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog

    dest = working_dir / "data"

    if star:
        src_path = f"{src_uri}/*"
    else:
        src_path = src_uri
        if cloud_type == "file":
            src_path += "/"

    if star:
        with pytest.raises(FileNotFoundError):
            catalog.cp([src_path], str(dest), recursive=recursive)

    if dir_exists or star:
        dest.mkdir()

    catalog.cp([src_path], str(dest), recursive=recursive)

    if not star and not recursive:
        # The root directory is skipped, so nothing is copied
        assert tree_from_path(dest) == {}
        return

    assert (dest / "description").read_text() == "Cats and Dogs"

    # Testing DVCx File Contents
    assert dest.with_suffix(".edvcx").is_file()
    edvcx_contents = yaml.safe_load(dest.with_suffix(".edvcx").read_text())
    assert len(edvcx_contents) == 1
    data = edvcx_contents[0]
    assert data["data-source"]["uri"] == src_path.rstrip("/")
    expected_file_count = 7 if recursive else 1
    assert len(data["files"]) == expected_file_count
    files_by_name = {f["name"]: f for f in data["files"]}

    # Directories should never be saved
    assert "cats" not in files_by_name
    assert "dogs" not in files_by_name
    assert "others" not in files_by_name
    assert "dogs/others" not in files_by_name

    # Description is always copied (if anything is copied)
    prefix = (
        "" if star or (recursive and not dir_exists) or cloud_type == "file" else "/"
    )
    assert files_by_name[f"{prefix}description"]["size"] == 13

    if recursive:
        assert tree_from_path(dest) == DEFAULT_TREE
        assert files_by_name[f"{prefix}cats/cat1"]["size"] == 4
        assert files_by_name[f"{prefix}cats/cat2"]["size"] == 4
        assert files_by_name[f"{prefix}dogs/dog1"]["size"] == 4
        assert files_by_name[f"{prefix}dogs/dog2"]["size"] == 3
        assert files_by_name[f"{prefix}dogs/dog3"]["size"] == 4
        assert files_by_name[f"{prefix}dogs/others/dog4"]["size"] == 4
        return

    assert (dest / "cats").exists() is False
    assert (dest / "dogs").exists() is False
    for prefix in ["/", ""]:
        assert f"{prefix}cats/cat1" not in files_by_name
        assert f"{prefix}cats/cat2" not in files_by_name
        assert f"{prefix}dogs/dog1" not in files_by_name
        assert f"{prefix}dogs/dog2" not in files_by_name
        assert f"{prefix}dogs/dog3" not in files_by_name
        assert f"{prefix}dogs/others/dog4" not in files_by_name


@pytest.mark.parametrize(
    "cloud_type",
    ["s3", "gcs", "azure"],
    indirect=True,
)
def test_cp_local_dataset(cloud_test_catalog, dogs_dataset):
    skip_if_not_sqlite()
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog

    dest = working_dir / "data"
    dest.mkdir()

    dataset_uri = dogs_dataset.uri(version=1)

    catalog.cp([dataset_uri], str(dest))

    parsed = urlparse(str(cloud_test_catalog.src))
    netloc = Path(parsed.netloc.strip("/"))
    path = Path(parsed.path.strip("/"))

    assert tree_from_path(dest / netloc / path) == {
        "dogs": {
            "dog1": "woof",
            "dog2": "arf",
            "dog3": "bark",
            "others": {"dog4": "ruff"},
        }
    }


@pytest.mark.parametrize("tree", [TARRED_TREE], indirect=True)
@pytest.mark.parametrize("suffix", ["/", "/*"])
@pytest.mark.parametrize("recursive", [False, True])
@pytest.mark.parametrize("dir_exists", [False, True])
@pytest.mark.xfail(reason="Missing support for v-objects in cp")
def test_cp_tar_root(cloud_test_catalog, suffix, recursive, dir_exists):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    create_tar_dataset(catalog, ctc.src_uri, "tarred")
    dest = ctc.working_dir / "data"
    if dir_exists:
        dest.mkdir()
    src = f"ds://tarred/animals.tar{suffix}"
    dest_path = str(dest) + "/"

    if not dir_exists and suffix == "/*":
        with pytest.raises(FileNotFoundError):
            catalog.cp([src], dest_path, recursive=recursive, no_edvcx_file=True)
        return

    catalog.cp([src], dest_path, recursive=recursive, no_edvcx_file=True)

    expected = DEFAULT_TREE.copy()
    if not recursive:
        # Directories are not copied
        if suffix == "/":
            expected = {}
        else:
            for key in list(expected):
                if isinstance(expected[key], dict):
                    del expected[key]

    assert tree_from_path(dest) == expected


@pytest.mark.parametrize("tree", [TARRED_TREE], indirect=True)
@pytest.mark.xfail(reason="Missing support for v-objects in cp")
def test_cp_full_tar(cloud_test_catalog):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    create_tar_dataset(catalog, ctc.src_uri, "tarred")
    dest = ctc.working_dir / "data"
    dest.mkdir()
    src = "ds://tarred/"
    catalog.cp([src], str(dest), recursive=True, no_edvcx_file=True)

    assert tree_from_path(dest, binary=True) == TARRED_TREE


@pytest.mark.parametrize(
    "recursive,star,slash,dir_exists",
    (
        (True, True, False, False),
        (True, False, False, False),
        (True, False, False, True),
        (True, False, True, False),
        (False, True, False, False),
        (False, False, False, False),
        (False, False, True, False),
    ),
)
def test_cp_subdir(cloud_test_catalog, recursive, star, slash, dir_exists):
    src_uri = f"{cloud_test_catalog.src_uri}/dogs"
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog

    dest = working_dir / "data"

    if star:
        src_path = f"{src_uri}/*"
    elif slash:
        src_path = f"{src_uri}/"
    else:
        src_path = src_uri

    if star:
        with pytest.raises(FileNotFoundError):
            catalog.cp([src_path], str(dest), recursive=recursive)

    if dir_exists or star:
        dest.mkdir()

    catalog.cp([src_path], str(dest), recursive=recursive)

    if not star and not recursive:
        # Directories are skipped, so nothing is copied
        assert tree_from_path(dest) == {}
        return

    # Testing DVCx File Contents
    assert dest.with_suffix(".edvcx").is_file()
    edvcx_contents = yaml.safe_load(dest.with_suffix(".edvcx").read_text())
    assert len(edvcx_contents) == 1
    data = edvcx_contents[0]
    assert data["data-source"]["uri"] == src_path.rstrip("/")
    expected_file_count = 4 if recursive else 3
    assert len(data["files"]) == expected_file_count
    files_by_name = {f["name"]: f for f in data["files"]}

    # Directories should never be saved
    assert "others" not in files_by_name
    assert "dogs/others" not in files_by_name

    if not dir_exists:
        assert (dest / "dog1").read_text() == "woof"
        assert (dest / "dog2").read_text() == "arf"
        assert (dest / "dog3").read_text() == "bark"
        assert (dest / "dogs").exists() is False
        assert files_by_name["dog1"]["size"] == 4
        assert files_by_name["dog2"]["size"] == 3
        assert files_by_name["dog3"]["size"] == 4
        if recursive:
            assert (dest / "others" / "dog4").read_text() == "ruff"
            assert files_by_name["others/dog4"]["size"] == 4
        else:
            assert (dest / "others").exists() is False
            assert "others/dog4" not in files_by_name
        return

    assert tree_from_path(dest / "dogs") == DEFAULT_TREE["dogs"]
    assert (dest / "dog1").exists() is False
    assert (dest / "dog2").exists() is False
    assert (dest / "dog3").exists() is False
    assert (dest / "others").exists() is False
    assert files_by_name["dogs/dog1"]["size"] == 4
    assert files_by_name["dogs/dog2"]["size"] == 3
    assert files_by_name["dogs/dog3"]["size"] == 4
    assert files_by_name["dogs/others/dog4"]["size"] == 4


@pytest.mark.parametrize("tree", [TARRED_TREE], indirect=True)
@pytest.mark.parametrize("path", ["*/dogs", "animals.tar/dogs"])
@pytest.mark.parametrize("suffix", ["", "/", "/*"])
@pytest.mark.parametrize("recursive", [False, True])
@pytest.mark.parametrize("dir_exists", [False, True])
@pytest.mark.xfail(reason="Missing support for v-objects in cp")
def test_cp_tar_subdir(cloud_test_catalog, path, suffix, recursive, dir_exists):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    create_tar_dataset(catalog, ctc.src_uri, "tarred")
    dest = ctc.working_dir / "data"
    if dir_exists:
        dest.mkdir()
    src = f"ds://tarred/{path}{suffix}"

    if not dir_exists and suffix == "/*":
        with pytest.raises(FileNotFoundError):
            catalog.cp([src], str(dest), recursive=recursive)
        return

    catalog.cp([src], str(dest), recursive=recursive)

    expected = DEFAULT_TREE["dogs"].copy()
    if suffix in ("",) and dir_exists:
        expected = {"dogs": expected}
    if not recursive:
        # Directories are not copied
        if not dir_exists or suffix == "/":
            expected = {}
        else:
            for key in list(expected):
                if isinstance(expected[key], dict):
                    del expected[key]

    assert tree_from_path(dest) == expected


@pytest.mark.parametrize(
    "recursive,star,slash",
    (
        (True, True, False),
        (True, False, False),
        (True, False, True),
        (False, True, False),
        (False, False, False),
        (False, False, True),
    ),
)
def test_cp_multi_subdir(cloud_test_catalog, recursive, star, slash):  # noqa: PLR0915
    sources = [
        f"{cloud_test_catalog.src_uri}/cats",
        f"{cloud_test_catalog.src_uri}/dogs",
    ]
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog

    dest = working_dir / "data"

    if star:
        src_paths = [f"{src}/*" for src in sources]
    elif slash:
        src_paths = [f"{src}/" for src in sources]
    else:
        src_paths = sources

    with pytest.raises(FileNotFoundError):
        catalog.cp(src_paths, str(dest), recursive=recursive)

    dest.mkdir()

    catalog.cp(src_paths, str(dest), recursive=recursive)

    if not star and not recursive:
        # Directories are skipped, so nothing is copied
        assert tree_from_path(dest) == {}
        return

    # Testing DVCx File Contents
    assert dest.with_suffix(".edvcx").is_file()
    edvcx_contents = yaml.safe_load(dest.with_suffix(".edvcx").read_text())
    assert len(edvcx_contents) == 2
    data_cats = edvcx_contents[0]
    data_dogs = edvcx_contents[1]
    assert data_cats["data-source"]["uri"] == src_paths[0].rstrip("/")
    assert data_dogs["data-source"]["uri"] == src_paths[1].rstrip("/")
    assert len(data_cats["files"]) == 2
    assert len(data_dogs["files"]) == 4 if recursive else 3
    cat_files_by_name = {f["name"]: f for f in data_cats["files"]}
    dog_files_by_name = {f["name"]: f for f in data_dogs["files"]}

    # Directories should never be saved
    assert "others" not in dog_files_by_name
    assert "dogs/others" not in dog_files_by_name

    if star or slash:
        assert (dest / "cat1").read_text() == "meow"
        assert (dest / "cat2").read_text() == "mrow"
        assert (dest / "dog1").read_text() == "woof"
        assert (dest / "dog2").read_text() == "arf"
        assert (dest / "dog3").read_text() == "bark"
        assert (dest / "cats").exists() is False
        assert (dest / "dogs").exists() is False
        assert cat_files_by_name["cat1"]["size"] == 4
        assert cat_files_by_name["cat2"]["size"] == 4
        assert dog_files_by_name["dog1"]["size"] == 4
        assert dog_files_by_name["dog2"]["size"] == 3
        assert dog_files_by_name["dog3"]["size"] == 4
        if recursive:
            assert (dest / "others" / "dog4").read_text() == "ruff"
            assert dog_files_by_name["others/dog4"]["size"] == 4
        else:
            assert (dest / "others").exists() is False
            assert "others/dog4" not in dog_files_by_name
        return

    assert (dest / "cats" / "cat1").read_text() == "meow"
    assert (dest / "cats" / "cat2").read_text() == "mrow"
    assert (dest / "dogs" / "dog1").read_text() == "woof"
    assert (dest / "dogs" / "dog2").read_text() == "arf"
    assert (dest / "dogs" / "dog3").read_text() == "bark"
    assert (dest / "dogs" / "others" / "dog4").read_text() == "ruff"
    assert (dest / "cat1").exists() is False
    assert (dest / "cat2").exists() is False
    assert (dest / "dog1").exists() is False
    assert (dest / "dog2").exists() is False
    assert (dest / "dog3").exists() is False
    assert (dest / "others").exists() is False
    assert cat_files_by_name["cats/cat1"]["size"] == 4
    assert cat_files_by_name["cats/cat2"]["size"] == 4
    assert dog_files_by_name["dogs/dog1"]["size"] == 4
    assert dog_files_by_name["dogs/dog2"]["size"] == 3
    assert dog_files_by_name["dogs/dog3"]["size"] == 4
    assert dog_files_by_name["dogs/others/dog4"]["size"] == 4


def test_cp_double_subdir(cloud_test_catalog):
    src_path = f"{cloud_test_catalog.src_uri}/dogs/others"
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog
    dest = working_dir / "data"

    catalog.cp([src_path], str(dest), recursive=True)

    # Testing DVCx File Contents
    assert dest.with_suffix(".edvcx").is_file()
    edvcx_contents = yaml.safe_load(dest.with_suffix(".edvcx").read_text())
    assert len(edvcx_contents) == 1
    data = edvcx_contents[0]
    assert data["data-source"]["uri"] == src_path.rstrip("/")
    assert len(data["files"]) == 1
    files_by_name = {f["name"]: f for f in data["files"]}

    # Directories should never be saved
    assert "others" not in files_by_name
    assert "dogs/others" not in files_by_name

    assert (dest / "dogs").exists() is False
    assert (dest / "others").exists() is False
    assert (dest / "dog4").read_text() == "ruff"
    assert files_by_name["dog4"]["size"] == 4


@pytest.mark.parametrize("no_glob", (True, False))
def test_cp_single_file(cloud_test_catalog, no_glob):
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog
    dest = working_dir / "data"
    src_path = f"{cloud_test_catalog.src_uri}/dogs/dog1"
    dest.mkdir()

    catalog.cp([src_path], str(dest / "local_dog"), no_edvcx_file=True, no_glob=no_glob)

    assert tree_from_path(dest) == {"local_dog": "woof"}


@pytest.mark.parametrize("tree", [{"foo": "original"}], indirect=True)
def test_storage_mutation(cloud_test_catalog):
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog
    src_path = f"{cloud_test_catalog.src_uri}/foo"

    dest = working_dir / "data1"
    dest.mkdir()
    catalog.cp([src_path], str(dest / "local"), no_edvcx_file=True)
    assert tree_from_path(dest) == {"local": "original"}

    # Storage modified without reindexing, we get the old version from cache.
    (cloud_test_catalog.src / "foo").write_text("modified")
    dest = working_dir / "data2"
    dest.mkdir()
    catalog.cp([src_path], str(dest / "local"), no_edvcx_file=True)
    assert tree_from_path(dest) == {"local": "original"}

    # Storage modified without reindexing.
    # Since the old version cannot be found in storage or cache, it's an error.
    catalog.cache.clear()
    dest = working_dir / "data3"
    dest.mkdir()
    with pytest.raises(FileNotFoundError):
        catalog.cp([src_path], str(dest / "local"), no_edvcx_file=True)
    assert tree_from_path(dest) == {}

    # Storage modified with reindexing, we get the new version.
    catalog.index([cloud_test_catalog.src_uri], update=True)
    dest = working_dir / "data4"
    dest.mkdir()
    catalog.cp([src_path], str(dest / "local"), no_edvcx_file=True)
    assert tree_from_path(dest) == {"local": "modified"}


def test_cp_edvcx_file_options(cloud_test_catalog):
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog
    dest = working_dir / "data"
    src_path = f"{cloud_test_catalog.src_uri}/dogs/*"
    edvcx_file = working_dir / "custom_name.edvcx"

    catalog.cp(
        [src_path],
        str(dest),
        recursive=False,
        edvcx_only=True,
        edvcx_file=str(edvcx_file),
    )

    assert (dest / "dog1").exists() is False
    assert (dest / "dog2").exists() is False
    assert (dest / "dog3").exists() is False
    assert (dest / "dogs").exists() is False
    assert (dest / "others").exists() is False
    assert dest.with_suffix(".edvcx").exists() is False

    # Testing DVCx File Contents
    assert edvcx_file.is_file()
    edvcx_contents = yaml.safe_load(edvcx_file.read_text())
    assert len(edvcx_contents) == 1
    data = edvcx_contents[0]
    assert data["data-source"]["uri"] == src_path
    expected_file_count = 3
    assert len(data["files"]) == expected_file_count
    files_by_name = {f["name"]: f for f in data["files"]}

    assert parse_edvcx_file(str(edvcx_file)) == edvcx_contents

    # Directories should never be saved
    assert "others" not in files_by_name
    assert "dogs/others" not in files_by_name

    assert files_by_name["dog1"]["size"] == 4
    assert files_by_name["dog2"]["size"] == 3
    assert files_by_name["dog3"]["size"] == 4
    assert "others/dog4" not in files_by_name

    with pytest.raises(FileNotFoundError):
        # Should fail, as * will not be expanded
        catalog.cp(
            [src_path],
            str(dest),
            recursive=False,
            edvcx_only=True,
            edvcx_file=str(edvcx_file),
            no_glob=True,
        )

    # Should succeed, as the DVCx file exists check will be skipped
    edvcx_only_data = catalog.cp(
        [src_path],
        str(dest),
        recursive=False,
        edvcx_only=True,
        edvcx_file=str(edvcx_file),
        force=True,
    )

    # Check the returned DVCx data contents
    assert len(edvcx_only_data) == len(edvcx_contents)
    edvcx_only_source = edvcx_only_data[0]
    assert edvcx_only_source["data-source"]["uri"] == src_path.rstrip("/")
    assert edvcx_only_source["files"] == data["files"]


def test_cp_edvcx_file_sources(cloud_test_catalog):  # noqa: PLR0915
    sources = [
        f"{cloud_test_catalog.src_uri}/cats/",
        f"{cloud_test_catalog.src_uri}/dogs/*",
    ]
    working_dir = cloud_test_catalog.working_dir
    catalog = cloud_test_catalog.catalog

    dest = working_dir / "data"

    edvcx_files = [
        working_dir / "custom_cats.edvcx",
        working_dir / "custom_dogs.edvcx",
    ]

    catalog.cp(
        sources[:1],
        str(dest),
        recursive=True,
        edvcx_only=True,
        edvcx_file=str(edvcx_files[0]),
    )

    catalog.cp(
        sources[1:],
        str(dest),
        recursive=True,
        edvcx_only=True,
        edvcx_file=str(edvcx_files[1]),
    )

    # Files should not be copied yet
    assert (dest / "cat1").exists() is False
    assert (dest / "cat2").exists() is False
    assert (dest / "cats").exists() is False
    assert (dest / "dog1").exists() is False
    assert (dest / "dog2").exists() is False
    assert (dest / "dog3").exists() is False
    assert (dest / "dogs").exists() is False
    assert (dest / "others").exists() is False

    # Testing DVCx File Contents
    edvcx_data = []
    for dqf in edvcx_files:
        assert dqf.is_file()
        edvcx_contents = yaml.safe_load(dqf.read_text())
        assert len(edvcx_contents) == 1
        edvcx_data.extend(edvcx_contents)

    assert len(edvcx_data) == 2
    data_cats1 = edvcx_data[0]
    data_dogs1 = edvcx_data[1]
    assert data_cats1["data-source"]["uri"] == sources[0].rstrip("/")
    assert data_dogs1["data-source"]["uri"] == sources[1].rstrip("/")
    assert len(data_cats1["files"]) == 2
    assert len(data_dogs1["files"]) == 4
    cat_files_by_name1 = {f["name"]: f for f in data_cats1["files"]}
    dog_files_by_name1 = {f["name"]: f for f in data_dogs1["files"]}

    # Directories should never be saved
    assert "others" not in dog_files_by_name1
    assert "dogs/others" not in dog_files_by_name1

    assert cat_files_by_name1["cat1"]["size"] == 4
    assert cat_files_by_name1["cat2"]["size"] == 4
    assert dog_files_by_name1["dog1"]["size"] == 4
    assert dog_files_by_name1["dog2"]["size"] == 3
    assert dog_files_by_name1["dog3"]["size"] == 4
    assert dog_files_by_name1["others/dog4"]["size"] == 4

    assert not dest.exists()

    with pytest.raises(FileNotFoundError):
        catalog.cp([str(dqf) for dqf in edvcx_files], str(dest), recursive=True)

    dest.mkdir()

    # Copy using these DVCx files as sources
    catalog.cp([str(dqf) for dqf in edvcx_files], str(dest), recursive=True)

    # Files should now be copied
    assert (dest / "cat1").read_text() == "meow"
    assert (dest / "cat2").read_text() == "mrow"
    assert (dest / "dog1").read_text() == "woof"
    assert (dest / "dog2").read_text() == "arf"
    assert (dest / "dog3").read_text() == "bark"
    assert (dest / "others" / "dog4").read_text() == "ruff"

    # Testing DVCx File Contents
    assert dest.with_suffix(".edvcx").is_file()
    edvcx_contents = yaml.safe_load(dest.with_suffix(".edvcx").read_text())
    assert len(edvcx_contents) == 2
    data_cats2 = edvcx_contents[0]
    data_dogs2 = edvcx_contents[1]
    assert data_cats2["data-source"]["uri"] == sources[0].rstrip("/")
    assert data_dogs2["data-source"]["uri"] == sources[1].rstrip("/")
    assert len(data_cats2["files"]) == 2
    assert len(data_dogs2["files"]) == 4
    cat_files_by_name2 = {f["name"]: f for f in data_cats2["files"]}
    dog_files_by_name2 = {f["name"]: f for f in data_dogs2["files"]}

    # Directories should never be saved
    assert "others" not in dog_files_by_name2
    assert "dogs/others" not in dog_files_by_name2

    assert cat_files_by_name2["cat1"]["size"] == 4
    assert cat_files_by_name2["cat2"]["size"] == 4
    assert dog_files_by_name2["dog1"]["size"] == 4
    assert dog_files_by_name2["dog2"]["size"] == 3
    assert dog_files_by_name2["dog3"]["size"] == 4
    assert dog_files_by_name2["others/dog4"]["size"] == 4


@pytest.mark.parametrize("cloud_type, version_aware", [("file", False)], indirect=True)
def test_cp_symlinks(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    catalog.client_config["use_symlinks"] = True
    src_uri = cloud_test_catalog.src_uri
    work_dir = cloud_test_catalog.working_dir
    dest = work_dir / "data"
    dest.mkdir()
    catalog.cp([f"{src_uri}/dogs/"], str(dest), recursive=True)

    assert (dest / "dog1").is_symlink()
    assert os.path.realpath(dest / "dog1") == str(
        cloud_test_catalog.src / "dogs" / "dog1"
    )
    assert (dest / "dog1").read_text() == "woof"
    assert (dest / "others" / "dog4").is_symlink()
    assert os.path.realpath(dest / "others" / "dog4") == str(
        cloud_test_catalog.src / "dogs" / "others" / "dog4"
    )
    assert (dest / "others" / "dog4").read_text() == "ruff"


def test_du(cloud_test_catalog):
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    expected_results = [
        (f"{src_uri}/cats/", 8),
        (f"{src_uri}/dogs/others/", 4),
        (f"{src_uri}/dogs/", 15),
        (f"{src_uri}/", 36),
    ]

    results = catalog.du([src_uri])
    assert set(results) == set(expected_results[3:])

    results = catalog.du([src_uri], depth=1)
    assert set(results) == set(expected_results[:1] + expected_results[2:])

    results = catalog.du([src_uri], depth=5)
    assert set(results) == set(expected_results)


def test_ls_glob(cloud_test_catalog):
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    assert sorted(
        (source.node.name, [r[0] for r in results])
        for source, results in catalog.ls([f"{src_uri}/dogs/dog*"], fields=["name"])
    ) == [("dog1", ["dog1"]), ("dog2", ["dog2"]), ("dog3", ["dog3"])]


def test_ls_prefix_not_found(cloud_test_catalog):
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog
    with pytest.raises(FileNotFoundError):
        list(catalog.ls([f"{src_uri}/bogus/"], fields=["name"]))


def clear_storages(catalog):
    ds = catalog.metastore
    ds.db.execute(ds._storages.delete())


@pytest.mark.parametrize("tree", [TARRED_TREE], indirect=True)
@pytest.mark.xfail(reason="Missing support for datasets in ls")
def test_ls_subobjects(cloud_test_catalog):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    create_tar_dataset(catalog, ctc.src_uri, "tarred")

    def do_ls(target):
        ((_, results),) = list(catalog.ls([target], fields=["name"]))
        results = list(results)
        result_set = {x[0] for x in results}
        assert len(result_set) == len(results)
        return result_set

    ds = "ds://tarred"
    assert do_ls(ds) == {"animals.tar"}
    assert do_ls(f"{ds}/animals.tar") == {"animals.tar"}
    assert do_ls(f"{ds}/animals.tar/dogs") == {
        "dog1",
        "dog2",
        "dog3",
        "others",
    }
    assert do_ls(f"{ds}/animals.tar/") == {"description", "cats", "dogs"}
    assert do_ls(f"{ds}/*.tar/") == {"description", "cats", "dogs"}
    assert do_ls(f"{ds}/*.tar/desc*") == {"description"}


def test_index_error(cloud_test_catalog):
    protocol = cloud_test_catalog.src_uri.split("://", 1)[0]
    # XXX: different clients raise inconsistent exceptions
    with pytest.raises(Exception):  # noqa: B017
        cloud_test_catalog.catalog.index([f"{protocol}://does_not_exist"])


def test_query(cloud_test_catalog, mock_popen_dataset_created):
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery
    DatasetQuery({src_uri!r})
    """
    query_script = dedent(query_script)

    dataset, version, query_output, _ = catalog.query(query_script, save=True)
    assert dataset
    assert_row_names(
        catalog,
        dataset,
        version,
        {
            "dog1",
            "dog2",
            "dog3",
            "dog4",
        },
    )
    assert dataset.query_script == query_script
    assert dataset.sources == ""


def test_query_save_size(cloud_test_catalog, mock_popen_dataset_created):
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    query_script = f"""\
    from dvcx.query import C, DatasetQuery
    DatasetQuery({src_uri!r})
    """
    query_script = dedent(query_script)

    dataset, version, _, _ = catalog.query(query_script, save=True)
    dataset_version = dataset.get_version(version)
    assert dataset_version.num_objects == 4
    assert dataset_version.size == 15


def test_query_fail_to_compile(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog

    query_script = "syntax error"

    with pytest.raises(QueryScriptCompileError):
        catalog.query(query_script)


def test_query_fail_wrong_dataset_name(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog

    query_script = """\
    from dvcx.query import DatasetQuery
    DatasetQuery("s3://bucket-name")
    """
    query_script = dedent(query_script)

    with pytest.raises(
        ValueError, match="Cannot use ds_query_ prefix for dataset name"
    ):
        catalog.query(query_script, save_as="ds_query_dataset")


def test_query_subprocess_wrong_return_code(mock_popen, cloud_test_catalog):
    mock_popen.configure_mock(returncode=1)
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    query_script = f"""
from dvcx.query import DatasetQuery, C
DatasetQuery('{src_uri}')
    """

    with pytest.raises(QueryScriptRunError) as exc_info:
        catalog.query(query_script)
        assert str(exc_info.value).startswith("Query script exited with error code 1")


def test_query_last_statement_not_expression(mock_popen, cloud_test_catalog):
    mock_popen.configure_mock(returncode=10)
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    query_script = f"""
from dvcx.query import DatasetQuery, C
ds = DatasetQuery('{src_uri}')
    """

    with pytest.raises(QueryScriptCompileError) as exc_info:
        catalog.query(query_script)
        assert str(exc_info.value).startswith(
            "Query script failed to compile, "
            "reason: Last line in a script was not an expression"
        )


def test_query_last_statement_not_ds_query_instance(mock_popen, cloud_test_catalog):
    mock_popen.configure_mock(returncode=10)
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    query_script = f"""
from dvcx.query import DatasetQuery, C
ds = DatasetQuery('{src_uri}')
5
    """

    with pytest.raises(QueryScriptRunError) as exc_info:
        catalog.query(query_script)
        assert str(exc_info.value).startswith(
            "Last line in a script was not an instance of DatasetQuery"
        )


def test_query_dataset_not_returned(mock_popen, cloud_test_catalog):
    mock_popen.configure_mock(stdout=io.StringIO("random str"))
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    query_script = f"""
from dvcx.query import DatasetQuery, C
DatasetQuery('{src_uri}')
    """

    with pytest.raises(QueryScriptDatasetNotFound) as e:
        catalog.query(query_script, save=True)
    assert e.value.output == "random str"


@pytest.mark.parametrize("cloud_type", ["s3", "azure", "gcs"], indirect=True)
def test_storage_stats(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    src_uri = cloud_test_catalog.src_uri

    with pytest.raises(StorageNotFoundError):
        catalog.storage_stats(src_uri)

    catalog.enlist_source(src_uri, ttl=1234)
    stats = catalog.storage_stats(src_uri)
    assert stats.num_objects == 11
    assert stats.size == 36

    catalog.enlist_source(f"{src_uri}/dogs/", ttl=1234, force_update=True)
    stats = catalog.storage_stats(src_uri)
    assert stats.num_objects == 6
    assert stats.size == 15


@pytest.mark.parametrize("from_cli", [False, True])
def test_garbage_collect(cloud_test_catalog, from_cli, capsys):
    catalog = cloud_test_catalog.catalog
    assert catalog.get_temp_table_names() == []
    temp_tables = ["tmp_vc12F", "udf_jh653", "ds_shadow_12345", "old_ds_shadow"]
    for t in temp_tables:
        catalog.warehouse.create_udf_table(t)
    assert set(catalog.get_temp_table_names()) == set(temp_tables)
    if from_cli:
        garbage_collect(catalog)
        captured = capsys.readouterr()
        assert captured.out == "Garbage collecting 4 tables.\n"
    else:
        catalog.cleanup_temp_tables(temp_tables)
    assert catalog.get_temp_table_names() == []



import copy
import hashlib
import json
import logging
import posixpath
from abc import ABC, abstractmethod
from collections.abc import Iterator
from datetime import datetime, timezone
from functools import cached_property, reduce
from itertools import groupby
from typing import TYPE_CHECKING, Any, Optional

from sqlalchemy import (
    JSON,
    BigInteger,
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    Table,
    Text,
    UniqueConstraint,
    select,
)
from sqlalchemy.sql import func

from dvcx.data_storage.serializer import Serializable
from dvcx.dataset import DatasetDependency, DatasetRecord, DatasetVersion
from dvcx.dataset import Status as DatasetStatus
from dvcx.error import DatasetNotFoundError, StorageNotFoundError, TableMissingError
from dvcx.sql.types import SQLType
from dvcx.storage import Status as StorageStatus
from dvcx.storage import Storage, StorageURI
from dvcx.utils import is_expired

if TYPE_CHECKING:
    from sqlalchemy.schema import SchemaItem

    from dvcx.data_storage import AbstractIDGenerator, schema
    from dvcx.data_storage.db_engine import DatabaseEngine


logger = logging.getLogger("dvcx")


class AbstractMetastore(ABC, Serializable):
    """
    Abstract Metastore class.
    This manages the storing, searching, and retrieval of indexed metadata.
    """

    uri: StorageURI
    partial_id: Optional[int]

    schema: "schema.Schema"
    storage_class: type[Storage] = Storage
    dataset_class: type[DatasetRecord] = DatasetRecord
    dependency_class: type[DatasetDependency] = DatasetDependency

    def __init__(
        self,
        uri: StorageURI = StorageURI(""),
        partial_id: Optional[int] = None,
    ):
        self.uri = uri
        self.partial_id: Optional[int] = partial_id

    @abstractmethod
    def clone(
        self,
        uri: StorageURI = StorageURI(""),
        partial_id: Optional[int] = None,
        use_new_connection: bool = False,
    ) -> "AbstractMetastore":
        """Clones AbstractMetastore implementation for some Storage input.
        Setting use_new_connection will always use a new database connection.
        New connections should only be used if needed due to errors with
        closed connections."""

    def init(self, uri: StorageURI, partial_id: int) -> None:
        """Initialize metastore."""

    def close(self) -> None:
        """Closes any active database or HTTP connections."""

    def cleanup_temp_tables(self, temp_table_names: list[str]):
        """Cleanup temp tables."""

    def cleanup_for_tests(self):
        """Cleanup for tests."""

    #
    # Storages
    #

    @abstractmethod
    def create_storage_if_not_registered(self, uri: StorageURI) -> None:
        """Saves new storage if it doesn't exist in database."""

    @abstractmethod
    def register_storage_for_indexing(
        self,
        uri: StorageURI,
        force_update: bool = True,
        prefix: str = "",
    ) -> tuple[Storage, bool, bool, Optional[int], Optional[str]]:
        """
        Prepares storage for indexing operation.
        This method should be called before index operation is started
        It returns:
            - storage, prepared for indexing
            - boolean saying if indexing is needed
            - boolean saying if indexing is currently pending (running)
            - partial id
            - partial path
        """

    @abstractmethod
    def find_stale_storages(self) -> None:
        """
        Finds all pending storages for which the last inserted node has happened
        before STALE_MINUTES_LIMIT minutes, and marks it as STALE.
        """

    @abstractmethod
    def mark_storage_indexed(
        self,
        uri: StorageURI,
        status: int,
        ttl: int,
        end_time: Optional[datetime] = None,
        prefix: str = "",
        partial_id: int = 0,
        error_message: str = "",
        error_stack: str = "",
        dataset: Optional[DatasetRecord] = None,
    ) -> None:
        """
        Marks storage as indexed.
        This method should be called when index operation is finished.
        """

    @abstractmethod
    def mark_storage_not_indexed(self, uri: StorageURI) -> None:
        """
        Mark storage as not indexed.
        This method should be called when storage index is deleted.
        """

    @abstractmethod
    def update_last_inserted_at(self, uri: Optional[StorageURI] = None) -> None:
        """Updates last inserted datetime in bucket with current time."""

    @abstractmethod
    def get_all_storage_uris(self) -> Iterator[StorageURI]:
        """Returns all storage uris."""

    @abstractmethod
    def get_storage(self, uri: StorageURI) -> Storage:
        """
        Gets storage representation from database.
        E.g. if s3 is used as storage this would be s3 bucket data.
        """

    @abstractmethod
    def list_storages(self) -> list[Storage]:
        """Returns all storages."""

    @abstractmethod
    def mark_storage_pending(self, storage: Storage) -> Storage:
        """Marks storage as pending."""

    #
    # Partial Indexes
    #

    @abstractmethod
    def init_partial_id(self, uri: StorageURI) -> None:
        """Initializes partial id for given storage."""

    @abstractmethod
    def get_next_partial_id(self, uri: StorageURI) -> int:
        """Returns next partial id for given storage."""

    @abstractmethod
    def get_valid_partial_id(
        self, uri: StorageURI, prefix: str, raise_exc: bool = True
    ) -> tuple[Optional[int], Optional[str]]:
        """
        Returns valid partial id and it's path, if they exist, for a given storage.
        """

    @abstractmethod
    def get_last_partial_path(self, uri: StorageURI) -> Optional[str]:
        """Returns last partial path for given storage."""

    #
    # Datasets
    #

    @abstractmethod
    def create_dataset(
        self,
        name: str,
        status: int = DatasetStatus.CREATED,
        sources: Optional[list[str]] = None,
        query_script: str = "",
        custom_column_types: Optional[dict[str, Any]] = None,
        ignore_if_exists: bool = False,
    ) -> DatasetRecord:
        """Creates new dataset."""

    @abstractmethod
    def create_dataset_version(
        self,
        dataset: DatasetRecord,
        version: int,
        status: int = DatasetStatus.CREATED,
        sources: str = "",
        query_script: str = "",
        error_message: str = "",
        error_stack: str = "",
        script_output: str = "",
        created_at: Optional[datetime] = None,
        finished_at: Optional[datetime] = None,
        custom_column_types: Optional[dict[str, Any]] = None,
        ignore_if_exists: bool = False,
        num_objects: Optional[int] = None,
        size: Optional[int] = None,
        preview: Optional[list[dict]] = None,
    ) -> DatasetRecord:
        """Creates new dataset version."""

    @abstractmethod
    def remove_dataset(self, dataset: DatasetRecord) -> None:
        """Removes dataset."""

    @abstractmethod
    def update_dataset(self, dataset: DatasetRecord, **kwargs) -> DatasetRecord:
        """Updates dataset fields."""

    @abstractmethod
    def update_dataset_version(
        self, dataset: DatasetRecord, version: int, **kwargs
    ) -> DatasetVersion:
        """Updates dataset version fields."""

    @abstractmethod
    def remove_dataset_version(
        self, dataset: DatasetRecord, version: int
    ) -> DatasetRecord:
        """
        Deletes one single dataset version.
        If it was last version, it removes dataset completely.
        """

    @abstractmethod
    def list_datasets(self) -> Iterator[DatasetRecord]:
        """Lists all datasets."""

    @abstractmethod
    def list_datasets_by_prefix(self, prefix: str) -> Iterator["DatasetRecord"]:
        """Lists all datasets which names start with prefix."""

    @abstractmethod
    def get_dataset(self, name: str) -> DatasetRecord:
        """Gets a single dataset by name."""

    @abstractmethod
    def update_dataset_status(
        self,
        dataset: DatasetRecord,
        status: int,
        version: Optional[int] = None,
        error_message="",
        error_stack="",
        script_output="",
    ) -> DatasetRecord:
        """Updates dataset status and appropriate fields related to status."""

    #
    # Dataset dependencies
    #

    def add_dependency(
        self,
        dependency: DatasetDependency,
        source_dataset_name: str,
        source_dataset_version: int,
    ) -> None:
        """Add dependency to dataset or storage."""
        if dependency.is_dataset:
            self.add_dataset_dependency(
                source_dataset_name,
                source_dataset_version,
                dependency.name,
                int(dependency.version),
            )
        else:
            self.add_storage_dependency(
                source_dataset_name,
                source_dataset_version,
                StorageURI(dependency.name),
                dependency.version,
            )

    @abstractmethod
    def add_storage_dependency(
        self,
        source_dataset_name: str,
        source_dataset_version: int,
        storage_uri: StorageURI,
        storage_timestamp_str: Optional[str] = None,
    ) -> None:
        """Adds storage dependency to dataset."""

    @abstractmethod
    def add_dataset_dependency(
        self,
        source_dataset_name: str,
        source_dataset_version: int,
        dataset_name: str,
        dataset_version: int,
    ) -> None:
        """Adds dataset dependency to dataset."""

    @abstractmethod
    def update_dataset_dependency_source(
        self,
        source_dataset: DatasetRecord,
        source_dataset_version: int,
        new_source_dataset: Optional[DatasetRecord] = None,
        new_source_dataset_version: Optional[int] = None,
    ) -> None:
        """Updates dataset dependency source."""

    @abstractmethod
    def get_direct_dataset_dependencies(
        self, dataset: DatasetRecord, version: int
    ) -> list[Optional[DatasetDependency]]:
        """Gets direct dataset dependencies."""

    @abstractmethod
    def remove_dataset_dependencies(
        self, dataset: DatasetRecord, version: Optional[int] = None
    ) -> None:
        """
        When we remove dataset, we need to clean up it's dependencies as well.
        """

    @abstractmethod
    def remove_dataset_dependants(
        self, dataset: DatasetRecord, version: Optional[int] = None
    ) -> None:
        """
        When we remove dataset, we need to clear its references in other dataset
        dependencies.
        """


class AbstractDBMetastore(AbstractMetastore):
    """
    Abstract Database Metastore class, to be implemented
    by any Database Adapters for a specific database system.
    This manages the storing, searching, and retrieval of indexed metadata,
    and has shared logic for all database systems currently in use.
    """

    PARTIALS_TABLE_NAME_PREFIX = "prt_"
    STORAGE_TABLE = "buckets"
    DATASET_TABLE = "datasets"
    DATASET_VERSION_TABLE = "datasets_versions"
    DATASET_DEPENDENCY_TABLE = "datasets_dependencies"

    id_generator: "AbstractIDGenerator"
    db: "DatabaseEngine"

    def __init__(
        self,
        id_generator: "AbstractIDGenerator",
        uri: StorageURI = StorageURI(""),
        partial_id: Optional[int] = None,
    ):
        self.id_generator = id_generator
        super().__init__(uri, partial_id)

    def close(self) -> None:
        """Closes any active database connections."""
        self.db.close()

    def cleanup_temp_tables(self, temp_table_names: list[str]):
        """Cleanup temp tables."""
        self.id_generator.delete_uris(temp_table_names)

    @classmethod
    def _buckets_columns(cls) -> list["SchemaItem"]:
        """Buckets (storages) table columns."""
        return [
            Column("id", Integer, primary_key=True, nullable=False),
            Column("uri", Text, nullable=False),
            Column("timestamp", DateTime(timezone=True)),
            Column("expires", DateTime(timezone=True)),
            Column("started_inserting_at", DateTime(timezone=True)),
            Column("last_inserted_at", DateTime(timezone=True)),
            Column("status", Integer, nullable=False),
            Column("error_message", Text, nullable=False, default=""),
            Column("error_stack", Text, nullable=False, default=""),
        ]

    @classmethod
    def _datasets_columns(cls) -> list["SchemaItem"]:
        """Datasets table columns."""
        return [
            Column("id", Integer, primary_key=True),
            Column("name", Text, nullable=False),
            Column("description", Text),
            Column("labels", JSON, nullable=True),
            Column("shadow", Boolean, nullable=False),
            Column("status", Integer, nullable=False),
            Column("created_at", DateTime(timezone=True)),
            Column("finished_at", DateTime(timezone=True)),
            Column("error_message", Text, nullable=False, default=""),
            Column("error_stack", Text, nullable=False, default=""),
            Column("script_output", Text, nullable=False, default=""),
            Column("sources", Text, nullable=False, default=""),
            Column("query_script", Text, nullable=False, default=""),
            Column("custom_column_types", JSON, nullable=True),
        ]

    @cached_property
    def _dataset_fields(self) -> list[str]:
        return [
            c.name  # type: ignore [attr-defined]
            for c in self._datasets_columns()
            if c.name  # type: ignore [attr-defined]
        ]

    @classmethod
    def _datasets_versions_columns(cls) -> list["SchemaItem"]:
        """Datasets versions table columns."""
        return [
            Column("id", Integer, primary_key=True),
            Column(
                "dataset_id",
                Integer,
                ForeignKey(f"{cls.DATASET_TABLE}.id", ondelete="CASCADE"),
                nullable=False,
            ),
            Column("version", Integer, nullable=False),
            # adding default for now until we fully remove shadow datasets
            Column("status", Integer, nullable=False, default=DatasetStatus.COMPLETE),
            Column("created_at", DateTime(timezone=True)),
            Column("finished_at", DateTime(timezone=True)),
            Column("error_message", Text, nullable=False, default=""),
            Column("error_stack", Text, nullable=False, default=""),
            Column("script_output", Text, nullable=False, default=""),
            Column("num_objects", BigInteger, nullable=True),
            Column("size", BigInteger, nullable=True),
            Column("preview", JSON, nullable=True),
            Column("sources", Text, nullable=False, default=""),
            Column("query_script", Text, nullable=False, default=""),
            Column("custom_column_types", JSON, nullable=True),
            UniqueConstraint("dataset_id", "version"),
        ]

    @cached_property
    def _dataset_version_fields(self) -> list[str]:
        return [
            c.name  # type: ignore [attr-defined]
            for c in self._datasets_versions_columns()
            if c.name  # type: ignore [attr-defined]
        ]

    @classmethod
    def _datasets_dependencies_columns(cls) -> list["SchemaItem"]:
        """Datasets dependencies table columns."""
        return [
            Column("id", Integer, primary_key=True),
            # TODO remove when https://github.com/iterative/dvcx/issues/959 is done
            Column(
                "source_dataset_id",
                Integer,
                ForeignKey(f"{cls.DATASET_TABLE}.id"),
                nullable=False,
            ),
            Column(
                "source_dataset_version_id",
                Integer,
                ForeignKey(f"{cls.DATASET_VERSION_TABLE}.id"),
                nullable=True,
            ),
            # TODO remove when https://github.com/iterative/dvcx/issues/959 is done
            Column(
                "dataset_id",
                Integer,
                ForeignKey(f"{cls.DATASET_TABLE}.id"),
                nullable=True,
            ),
            Column(
                "dataset_version_id",
                Integer,
                ForeignKey(f"{cls.DATASET_VERSION_TABLE}.id"),
                nullable=True,
            ),
            # TODO remove when https://github.com/iterative/dvcx/issues/1121 is done
            # If we unify datasets and bucket listing then both bucket fields won't
            # be needed
            Column(
                "bucket_id",
                Integer,
                ForeignKey(f"{cls.STORAGE_TABLE}.id"),
                nullable=True,
            ),
            Column("bucket_version", Text, nullable=True),
        ]

    @classmethod
    def _storage_partial_columns(cls) -> list["SchemaItem"]:
        """Storage partial table columns."""
        return [
            Column("path_str", Text, nullable=False),
            # This is generated before insert and is not the SQLite rowid,
            # so it is not the primary key.
            Column("partial_id", Integer, nullable=False, index=True),
            Column("timestamp", DateTime(timezone=True)),
            Column("expires", DateTime(timezone=True)),
        ]

    def _get_storage_partial_table(self, name: str) -> Table:
        table = self.db.metadata.tables.get(name)
        if table is None:
            table = Table(
                name,
                self.db.metadata,
                *self._storage_partial_columns(),
            )
        return table

    #
    # Query Tables
    #

    def _partials_table(self, uri: StorageURI) -> Table:
        return self._get_storage_partial_table(self._partials_table_name(uri))

    @cached_property
    def _storages(self) -> Table:
        return Table(self.STORAGE_TABLE, self.db.metadata, *self._buckets_columns())

    @cached_property
    def _partials(self) -> Table:
        assert (
            self._current_partials_table_name
        ), "Partials can only be used if uri/current_partials_table_name is set"
        return self._get_storage_partial_table(self._current_partials_table_name)

    @cached_property
    def _datasets(self) -> Table:
        return Table(self.DATASET_TABLE, self.db.metadata, *self._datasets_columns())

    @cached_property
    def _datasets_versions(self) -> Table:
        return Table(
            self.DATASET_VERSION_TABLE,
            self.db.metadata,
            *self._datasets_versions_columns(),
        )

    @cached_property
    def _datasets_dependencies(self) -> Table:
        return Table(
            self.DATASET_DEPENDENCY_TABLE,
            self.db.metadata,
            *self._datasets_dependencies_columns(),
        )

    #
    # Query Starters (These can be overridden by subclasses)
    #

    @abstractmethod
    def _storages_insert(self): ...

    def _storages_select(self, *columns):
        if not columns:
            return self._storages.select()
        return select(*columns).select_from(self._storages)

    def _storages_update(self):
        return self._storages.update()

    def _storages_delete(self):
        return self._storages.delete()

    @abstractmethod
    def _partials_insert(self): ...

    def _partials_select(self, *columns):
        if not columns:
            return self._partials.select()
        return select(*columns).select_from(self._partials)

    def _partials_update(self):
        return self._partials.update()

    @abstractmethod
    def _datasets_insert(self): ...

    def _datasets_select(self, *columns):
        if not columns:
            return self._datasets.select()
        return select(*columns).select_from(self._datasets)

    def _datasets_update(self):
        return self._datasets.update()

    def _datasets_delete(self):
        return self._datasets.delete()

    @abstractmethod
    def _datasets_versions_insert(self): ...

    def _datasets_versions_select(self, *columns):
        if not columns:
            return self._datasets_versions.select()
        return select(*columns).select_from(self._datasets_versions)

    def _datasets_versions_update(self):
        return self._datasets_versions.update()

    def _datasets_versions_delete(self):
        return self._datasets_versions.delete()

    @abstractmethod
    def _datasets_dependencies_insert(self): ...

    def _datasets_dependencies_select(self, *columns):
        if not columns:
            return self._datasets_dependencies.select()
        return select(*columns).select_from(self._datasets_dependencies)

    def _datasets_dependencies_update(self):
        return self._datasets_dependencies.update()

    def _datasets_dependencies_delete(self):
        return self._datasets_dependencies.delete()

    #
    # Table Name Internal Functions
    #

    def _partials_table_name(self, uri: StorageURI) -> str:
        sha = hashlib.sha256(uri.encode("utf-8")).hexdigest()[:12]
        return f"{self.PARTIALS_TABLE_NAME_PREFIX}_{sha}"

    @property
    def _current_partials_table_name(self) -> Optional[str]:
        if not self.uri:
            return None
        return self._partials_table_name(self.uri)

    #
    # Storages
    #

    def create_storage_if_not_registered(self, uri: StorageURI, conn=None) -> None:
        """Saves new storage if it doesn't exist in database."""
        self.db.execute(
            self._storages_insert()
            .values(
                uri=uri,
                status=StorageStatus.CREATED,
                error_message="",
                error_stack="",
            )
            .on_conflict_do_nothing(),
            conn=conn,
        )

    def register_storage_for_indexing(
        self,
        uri: StorageURI,
        force_update: bool = True,
        prefix: str = "",
    ) -> tuple[Storage, bool, bool, Optional[int], Optional[str]]:
        """
        Prepares storage for indexing operation.
        This method should be called before index operation is started
        It returns:
            - storage, prepared for indexing
            - boolean saying if indexing is needed
            - boolean saying if indexing is currently pending (running)
            - partial id
            - partial path
        """
        # This ensures that all calls to the DB are in a single transaction
        # and commit is automatically called once this function returns
        with self.db.transaction() as conn:
            # Create storage if it doesn't exist
            self.create_storage_if_not_registered(uri, conn=conn)
            storage = self.get_storage(uri, conn=conn)

            if storage.status == StorageStatus.PENDING:
                return storage, False, True, None, None

            if storage.is_expired or storage.status == StorageStatus.STALE:
                storage = self.mark_storage_pending(storage, conn=conn)
                return storage, True, False, None, None

            if (
                storage.status in (StorageStatus.PARTIAL, StorageStatus.COMPLETE)
                and not force_update
            ):
                partial_id, partial_path = self.get_valid_partial_id(
                    uri, prefix, raise_exc=False
                )
                if partial_id is not None:
                    return storage, False, False, partial_id, partial_path
                return storage, True, False, None, None

            storage = self.mark_storage_pending(storage, conn=conn)
            return storage, True, False, None, None

    def find_stale_storages(self) -> None:
        """
        Finds all pending storages for which the last inserted node has happened
        before STALE_MINUTES_LIMIT minutes, and marks it as STALE.
        """
        s = self._storages
        with self.db.transaction() as conn:
            pending_storages = map(
                self.storage_class._make,
                self.db.execute(
                    self._storages_select().where(s.c.status == StorageStatus.PENDING),
                    conn=conn,
                ),
            )
            for storage in pending_storages:
                if storage.is_stale:
                    print(f"Marking storage {storage.uri} as stale")
                    self._mark_storage_stale(storage.id, conn=conn)

    def mark_storage_indexed(
        self,
        uri: StorageURI,
        status: int,
        ttl: int,
        end_time: Optional[datetime] = None,
        prefix: str = "",
        partial_id: int = 0,
        error_message: str = "",
        error_stack: str = "",
        dataset: Optional[DatasetRecord] = None,
    ) -> None:
        """
        Marks storage as indexed.
        This method should be called when index operation is finished.
        """
        if status == StorageStatus.PARTIAL and not prefix:
            raise AssertionError("Partial indexing requires a prefix")

        if end_time is None:
            end_time = datetime.now(timezone.utc)
        expires = Storage.get_expiration_time(end_time, ttl)

        s = self._storages
        with self.db.transaction() as conn:
            self.db.execute(
                self._storages_update()
                .where(s.c.uri == uri)
                .values(  # type: ignore [attr-defined]
                    timestamp=end_time,
                    expires=expires,
                    status=status,
                    last_inserted_at=end_time,
                    error_message=error_message,
                    error_stack=error_stack,
                ),
                conn=conn,
            )

            if not self._current_partials_table_name:
                # This only occurs in tests
                return

            if status in (StorageStatus.PARTIAL, StorageStatus.COMPLETE):
                dir_prefix = posixpath.join(prefix, "")
                self.db.execute(
                    self._partials_insert().values(
                        path_str=dir_prefix,
                        timestamp=end_time,
                        expires=expires,
                        partial_id=partial_id,
                    ),
                    conn=conn,
                )

            # update underlying dataset status as well
            if status == StorageStatus.FAILED and dataset:
                self.update_dataset_status(
                    dataset,
                    DatasetStatus.FAILED,
                    dataset.latest_version,
                    error_message=error_message,
                    error_stack=error_stack,
                    conn=conn,
                )

            if status in (StorageStatus.PARTIAL, StorageStatus.COMPLETE) and dataset:
                self.update_dataset_status(
                    dataset, DatasetStatus.COMPLETE, dataset.latest_version, conn=conn
                )

    def update_last_inserted_at(self, uri: Optional[StorageURI] = None) -> None:
        """Updates last inserted datetime in bucket with current time"""
        uri = uri or self.uri
        updates = {"last_inserted_at": datetime.now(timezone.utc)}
        s = self._storages
        self.db.execute(
            self._storages_update().where(s.c.uri == uri).values(**updates)  # type: ignore [attr-defined]
        )

    def get_all_storage_uris(self) -> Iterator[StorageURI]:
        """Returns all storage uris."""
        s = self._storages
        yield from (r[0] for r in self.db.execute(self._storages_select(s.c.uri)))

    def get_storage(self, uri: StorageURI, conn=None) -> Storage:
        """
        Gets storage representation from database.
        E.g. if s3 is used as storage this would be s3 bucket data
        """
        s = self._storages
        result = next(
            self.db.execute(self._storages_select().where(s.c.uri == uri), conn=conn),
            None,
        )
        if not result:
            raise StorageNotFoundError(f"Storage {uri} not found.")

        return self.storage_class._make(result)

    def list_storages(self) -> list[Storage]:
        result = self.db.execute(self._storages_select())
        if not result:
            return []

        return [self.storage_class._make(r) for r in result]

    def mark_storage_pending(self, storage: Storage, conn=None) -> Storage:
        # Update status to pending and dates
        updates = {
            "status": StorageStatus.PENDING,
            "timestamp": None,
            "expires": None,
            "last_inserted_at": None,
            "started_inserting_at": datetime.now(timezone.utc),
        }
        storage = storage._replace(**updates)  # type: ignore [arg-type]
        s = self._storages
        self.db.execute(
            self._storages_update().where(s.c.uri == storage.uri).values(**updates),  # type: ignore [attr-defined]
            conn=conn,
        )
        return storage

    def _mark_storage_stale(self, storage_id: int, conn=None) -> None:
        # Update status to pending and dates
        updates = {"status": StorageStatus.STALE, "timestamp": None, "expires": None}
        s = self._storages
        self.db.execute(
            self._storages.update().where(s.c.id == storage_id).values(**updates),  # type: ignore [attr-defined]
            conn=conn,
        )

    #
    # Partial Indexes
    #

    def init_partial_id(self, uri: StorageURI) -> None:
        """Initializes partial id for given storage."""
        if not uri:
            raise ValueError("uri for get_next_partial_id() cannot be empty")
        self.id_generator.init_id(f"partials:{uri}")

    def get_next_partial_id(self, uri: StorageURI) -> int:
        """Returns next partial id for given storage."""
        if not uri:
            raise ValueError("uri for get_next_partial_id() cannot be empty")
        return self.id_generator.get_next_id(f"partials:{uri}")

    def get_valid_partial_id(
        self, uri: StorageURI, prefix: str, raise_exc: bool = True
    ) -> tuple[Optional[int], Optional[str]]:
        """
        Returns valid partial id and it's path, if they exist, for a given storage.
        """
        # This SQL statement finds all entries that are
        # prefixes of the given prefix, matching this or parent directories
        # that are indexed.
        dir_prefix = posixpath.join(prefix, "")
        p = self._partials_table(uri)
        expire_values = self.db.execute(
            select(p.c.expires, p.c.partial_id, p.c.path_str)
            .select_from(p)
            .where(
                p.c.path_str == func.substr(dir_prefix, 1, func.length(p.c.path_str))
            )
            .order_by(p.c.expires.desc())
        )
        for expires, partial_id, path_str in expire_values:
            if not is_expired(expires):
                return partial_id, path_str
        if raise_exc:
            raise RuntimeError(f"Unable to get valid partial_id: {uri=}, {prefix=}")
        return None, None

    def get_last_partial_path(self, uri: StorageURI) -> Optional[str]:
        """Returns last partial path for given storage."""
        p = self._partials_table(uri)
        if not self.db.has_table(p.name):
            raise StorageNotFoundError(f"Storage {uri} partials are not found.")
        last_partial = self.db.execute(
            select(p.c.path_str).select_from(p).order_by(p.c.timestamp.desc()).limit(1)
        )
        for (path_str,) in last_partial:
            return path_str
        return None

    #
    # Datasets
    #

    def create_dataset(
        self,
        name: str,
        status: int = DatasetStatus.CREATED,
        sources: Optional[list[str]] = None,
        query_script: str = "",
        custom_column_types: Optional[dict[str, Any]] = None,
        ignore_if_exists: bool = False,
        **kwargs,  # TODO registered = True / False
    ) -> DatasetRecord:
        """Creates new dataset."""
        column_types = {
            c.name: c.type.to_dict()
            for c in self.schema.dataset_row_cls.dataset_default_columns()
            if isinstance(c.type, SQLType)
        }
        if custom_column_types:
            column_types.update(custom_column_types)

        # TODO abstract this method and add registered = True based on kwargs
        query = self._datasets_insert().values(
            name=name,
            shadow=False,
            status=status,
            created_at=datetime.now(timezone.utc),
            error_message="",
            error_stack="",
            script_output="",
            sources="\n".join(sources) if sources else "",
            query_script=query_script,
            custom_column_types=json.dumps(column_types),
        )
        if ignore_if_exists:
            query = query.on_conflict_do_nothing(index_elements=["name"])
        self.db.execute(query)

        return self.get_dataset(name)

    def create_dataset_version(  # noqa: PLR0913
        self,
        dataset: DatasetRecord,
        version: int,
        status: int = DatasetStatus.CREATED,
        sources: str = "",
        query_script: str = "",
        error_message: str = "",
        error_stack: str = "",
        script_output: str = "",
        created_at: Optional[datetime] = None,
        finished_at: Optional[datetime] = None,
        custom_column_types: Optional[dict[str, Any]] = None,
        ignore_if_exists: bool = False,
        num_objects: Optional[int] = None,
        size: Optional[int] = None,
        preview: Optional[list[dict]] = None,
        conn=None,
    ) -> DatasetRecord:
        """Creates new dataset version."""
        if status in [DatasetStatus.COMPLETE, DatasetStatus.FAILED]:
            finished_at = finished_at or datetime.now(timezone.utc)
        else:
            finished_at = None

        query = self._datasets_versions_insert().values(
            dataset_id=dataset.id,
            version=version,
            status=status,  # for now until we remove shadow datasets
            created_at=created_at or datetime.now(timezone.utc),
            finished_at=finished_at,
            error_message=error_message,
            error_stack=error_stack,
            script_output=script_output,
            sources=sources,
            query_script=query_script,
            custom_column_types=json.dumps(custom_column_types or {}),
            num_objects=num_objects,
            size=size,
            preview=json.dumps(preview) if preview else None,
        )
        if ignore_if_exists:
            query = query.on_conflict_do_nothing(
                index_elements=["dataset_id", "version"]
            )
        self.db.execute(query, conn=conn)

        return self.get_dataset(dataset.name, conn=conn)

    def remove_dataset(self, dataset: DatasetRecord) -> None:
        """Removes dataset."""
        d = self._datasets
        with self.db.transaction():
            self.remove_dataset_dependencies(dataset)
            self.remove_dataset_dependants(dataset)
            self.db.execute(self._datasets_delete().where(d.c.id == dataset.id))

    def update_dataset(
        self, dataset: DatasetRecord, conn=None, **kwargs
    ) -> DatasetRecord:
        """Updates dataset fields."""
        values = {}
        dataset_values = {}
        for field, value in kwargs.items():
            if field in self._dataset_fields[1:]:
                if field in ["labels", "custom_column_types"]:
                    values[field] = json.dumps(value) if value else None
                else:
                    values[field] = value
                if field == "custom_column_types":
                    dataset_values[field] = DatasetRecord.parse_custom_column_types(
                        value
                    )
                else:
                    dataset_values[field] = value

        if not values:
            # Nothing to update
            return dataset

        d = self._datasets
        self.db.execute(
            self._datasets_update().where(d.c.name == dataset.name).values(values),
            conn=conn,
        )  # type: ignore [attr-defined]

        result_ds = copy.deepcopy(dataset)
        result_ds.update(**dataset_values)
        return result_ds

    def update_dataset_version(
        self, dataset: DatasetRecord, version: int, conn=None, **kwargs
    ) -> DatasetVersion:
        """Updates dataset fields."""
        dataset_version = dataset.get_version(version)

        values = {}
        for field, value in kwargs.items():
            if field in self._dataset_version_fields[1:]:
                if field == "custom_column_types":
                    dataset_version.update(
                        **{field: DatasetRecord.parse_custom_column_types(value)}
                    )
                    values[field] = json.dumps(value) if value else None
                elif field == "preview" and isinstance(value, list):
                    values[field] = json.dumps(value)
                else:
                    values[field] = value
                    dataset_version.update(**{field: value})

        if not values:
            # Nothing to update
            return dataset_version

        dv = self._datasets_versions
        self.db.execute(
            self._datasets_versions_update()
            .where(dv.c.id == dataset_version.id)
            .values(values),
            conn=conn,
        )  # type: ignore [attr-defined]

        return dataset_version

    def _parse_dataset(self, rows) -> Optional[DatasetRecord]:
        versions = []
        for r in rows:
            versions.append(self.dataset_class.parse(*r))
        if not versions:
            return None
        return reduce(lambda ds, version: ds.merge_versions(version), versions)

    def _parse_datasets(self, rows) -> Iterator["DatasetRecord"]:
        # grouping rows by dataset id
        for _, g in groupby(rows, lambda r: r[0]):
            dataset = self._parse_dataset(list(g))
            if dataset:
                yield dataset

    def _base_dataset_query(self):
        if not (
            self.db.has_table(self._datasets.name)
            and self.db.has_table(self._datasets_versions.name)
        ):
            raise TableMissingError

        d = self._datasets
        dv = self._datasets_versions
        query = self._datasets_select(
            *(getattr(d.c, f) for f in self._dataset_fields),
            *(getattr(dv.c, f) for f in self._dataset_version_fields),
        )
        j = d.join(dv, d.c.id == dv.c.dataset_id, isouter=True)
        return query.select_from(j)

    def list_datasets(self) -> Iterator["DatasetRecord"]:
        """Lists all datasets."""
        yield from self._parse_datasets(self.db.execute(self._base_dataset_query()))

    def list_datasets_by_prefix(
        self, prefix: str, conn=None
    ) -> Iterator["DatasetRecord"]:
        query = self._base_dataset_query()
        query = query.where(self._datasets.c.name.startswith(prefix))
        yield from self._parse_datasets(self.db.execute(query))

    def get_dataset(self, name: str, conn=None) -> DatasetRecord:
        """Gets a single dataset by name"""
        d = self._datasets
        query = self._base_dataset_query()
        query = query.where(d.c.name == name)  # type: ignore [attr-defined]
        ds = self._parse_dataset(self.db.execute(query, conn=conn))
        if not ds:
            raise DatasetNotFoundError(f"Dataset {name} not found.")
        return ds

    def remove_dataset_version(
        self, dataset: DatasetRecord, version: int
    ) -> DatasetRecord:
        """
        Deletes one single dataset version.
        If it was last version, it removes dataset completely
        """
        if not dataset.has_version(version):
            raise DatasetNotFoundError(
                f"Dataset {dataset.name} version {version} not found."
            )

        self.remove_dataset_dependencies(dataset, version)
        self.remove_dataset_dependants(dataset, version)

        d = self._datasets
        dv = self._datasets_versions
        self.db.execute(
            self._datasets_versions_delete().where(
                (dv.c.dataset_id == dataset.id) & (dv.c.version == version)
            )
        )

        if dataset.versions and len(dataset.versions) == 1:
            # had only one version, fully deleting dataset
            self.db.execute(self._datasets_delete().where(d.c.id == dataset.id))

        dataset.remove_version(version)
        return dataset

    def update_dataset_status(
        self,
        dataset: DatasetRecord,
        status: int,
        version: Optional[int] = None,
        error_message="",
        error_stack="",
        script_output="",
        conn=None,
    ) -> DatasetRecord:
        """
        Updates dataset status and appropriate fields related to status
        It also updates version if specified.
        """
        update_data: dict[str, Any] = {"status": status}
        if status in [DatasetStatus.COMPLETE, DatasetStatus.FAILED]:
            # if in final state, updating finished_at datetime
            update_data["finished_at"] = datetime.now(timezone.utc)
            if script_output:
                update_data["script_output"] = script_output

        if status == DatasetStatus.FAILED:
            update_data["error_message"] = error_message
            update_data["error_stack"] = error_stack

        self.update_dataset(dataset, conn=conn, **update_data)

        if version:
            self.update_dataset_version(dataset, version, conn=conn, **update_data)

        return dataset

    #
    # Dataset dependencies
    #

    def _insert_dataset_dependency(self, data: dict[str, Any]) -> None:
        """Method for inserting dependencies."""
        self.db.execute(self._datasets_dependencies_insert().values(**data))

    def add_storage_dependency(
        self,
        source_dataset_name: str,
        source_dataset_version: int,
        storage_uri: StorageURI,
        storage_timestamp_str: Optional[str] = None,
    ) -> None:
        source_dataset = self.get_dataset(source_dataset_name)
        storage = self.get_storage(storage_uri)

        self._insert_dataset_dependency(
            {
                "source_dataset_id": source_dataset.id,
                "source_dataset_version_id": (
                    source_dataset.get_version(source_dataset_version).id
                ),
                "bucket_id": storage.id,
                "bucket_version": storage_timestamp_str,
            }
        )

    def add_dataset_dependency(
        self,
        source_dataset_name: str,
        source_dataset_version: int,
        dataset_name: str,
        dataset_version: int,
    ) -> None:
        """Adds dataset dependency to dataset."""
        source_dataset = self.get_dataset(source_dataset_name)
        dataset = self.get_dataset(dataset_name)

        self._insert_dataset_dependency(
            {
                "source_dataset_id": source_dataset.id,
                "source_dataset_version_id": (
                    source_dataset.get_version(source_dataset_version).id
                ),
                "dataset_id": dataset.id,
                "dataset_version_id": dataset.get_version(dataset_version).id,
            }
        )

    def update_dataset_dependency_source(
        self,
        source_dataset: DatasetRecord,
        source_dataset_version: int,
        new_source_dataset: Optional[DatasetRecord] = None,
        new_source_dataset_version: Optional[int] = None,
    ) -> None:
        dd = self._datasets_dependencies

        if not new_source_dataset:
            new_source_dataset = source_dataset

        q = self._datasets_dependencies_update().where(
            dd.c.source_dataset_id == source_dataset.id
        )
        q = q.where(
            dd.c.source_dataset_version_id
            == source_dataset.get_version(source_dataset_version).id
        )

        data = {"source_dataset_id": new_source_dataset.id}
        if new_source_dataset_version:
            data["source_dataset_version_id"] = new_source_dataset.get_version(
                new_source_dataset_version
            ).id

        q = q.values(**data)
        self.db.execute(q)

    @abstractmethod
    def _dataset_dependencies_select_columns(self) -> list["SchemaItem"]:
        """
        Returns a list of columns to select in a query for fetching dataset dependencies
        """

    def get_direct_dataset_dependencies(
        self, dataset: DatasetRecord, version: int
    ) -> list[Optional[DatasetDependency]]:
        d = self._datasets
        dd = self._datasets_dependencies
        dv = self._datasets_versions
        s = self._storages

        dataset_version = dataset.get_version(version)

        select_cols = self._dataset_dependencies_select_columns()

        query = (
            self._datasets_dependencies_select(*select_cols)
            .where(
                (dd.c.source_dataset_id == dataset.id)
                & (dd.c.source_dataset_version_id == dataset_version.id)
            )
            .select_from(
                dd.join(d, dd.c.dataset_id == d.c.id, isouter=True)
                .join(s, dd.c.bucket_id == s.c.id, isouter=True)
                .join(dv, dd.c.dataset_version_id == dv.c.id, isouter=True)
            )
        )
        if version:
            dataset_version = dataset.get_version(version)
            query = query.where(dd.c.source_dataset_version_id == dataset_version.id)

        return [self.dependency_class.parse(*r) for r in self.db.execute(query)]

    def remove_dataset_dependencies(
        self, dataset: DatasetRecord, version: Optional[int] = None
    ) -> None:
        """
        When we remove dataset, we need to clean up it's dependencies as well
        """
        dd = self._datasets_dependencies

        q = self._datasets_dependencies_delete().where(
            dd.c.source_dataset_id == dataset.id
        )

        if version:
            q = q.where(
                dd.c.source_dataset_version_id == dataset.get_version(version).id
            )

        self.db.execute(q)

    def remove_dataset_dependants(
        self, dataset: DatasetRecord, version: Optional[int] = None
    ) -> None:
        """
        When we remove dataset, we need to clear its references in other dataset
        dependencies
        """
        dd = self._datasets_dependencies

        q = self._datasets_dependencies_update().where(dd.c.dataset_id == dataset.id)
        if version:
            q = q.where(dd.c.dataset_version_id == dataset.get_version(version).id)

        q = q.values(dataset_id=None, dataset_version_id=None)

        self.db.execute(q)



import base64
import pickle

import pytest

from dvcx.data_storage.serializer import Serializable, deserialize


class MySerializableInit(Serializable):
    def __init__(self, name, optional=None):
        self.name = name
        self.optional = optional

    def clone_params(self):
        return MySerializableInit, [self.name], {"optional": self.optional}

    def get_params(self):
        return self.name, self.optional


class MySerializableFunc(Serializable):
    def __init__(self, name, optional=None):
        self.name = name
        self.optional = optional

    @classmethod
    def from_params(cls, name, optional=None):
        return cls(name, optional=optional)

    def clone_params(self):
        return self.from_params, [self.name], {"optional": self.optional}

    def get_params(self):
        return self.name, self.optional


class MySerializableNoParams(Serializable):
    def clone_params(self):
        return MySerializableNoParams, [], {}


@pytest.mark.parametrize(
    "cls,call",
    [
        (MySerializableInit, MySerializableInit),
        (MySerializableFunc, MySerializableFunc.from_params),
    ],
)
@pytest.mark.parametrize(
    "name,optional",
    [
        (None, None),
        ("foo", None),
        (None, 12),
        ("bar", 24),
    ],
)
def test_serializable_init(cls, call, name, optional):
    obj = cls(name, optional=optional)
    assert obj.clone_params() == (call, [name], {"optional": optional})

    serialized = obj.serialize()
    assert serialized
    serialized_pickled = base64.b64decode(serialized.encode())
    assert serialized_pickled
    (f, args, kwargs) = pickle.loads(serialized_pickled)  # noqa: S301
    assert f == call
    assert args == [name]
    assert kwargs == {"optional": optional}

    obj2 = deserialize(serialized)
    assert isinstance(obj2, cls)
    assert obj2.name == name
    assert obj2.optional == optional
    assert obj2.get_params() == (name, optional)


def test_serializable_init_no_params():
    obj = MySerializableNoParams()
    assert obj.clone_params() == (MySerializableNoParams, [], {})

    serialized = obj.serialize()
    assert serialized
    serialized_pickled = base64.b64decode(serialized.encode())
    assert serialized_pickled
    (f, args, kwargs) = pickle.loads(serialized_pickled)  # noqa: S301
    assert f == MySerializableNoParams
    assert args == []
    assert kwargs == {}

    obj2 = deserialize(serialized)
    assert isinstance(obj2, MySerializableNoParams)



import re
from datetime import datetime
from struct import pack
from time import sleep
from typing import Any

import msgpack
import pytest
from sqlalchemy import select

from dvcx.cli import ls
from dvcx.client.local import FileClient
from tests.utils import uppercase_scheme


def same_lines(lines1, lines2):
    def _split_lines(lines):
        return [line.strip() for line in sorted(lines.split("\n"))]

    return _split_lines(lines1) == _split_lines(lines2)


def test_ls_no_args(cloud_test_catalog, cloud_type, capsys):
    src = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog
    catalog.index([src])
    ls([], catalog=catalog)
    captured = capsys.readouterr()
    if cloud_type == "file":
        assert captured.out == FileClient.root_path().as_uri() + "\n"
    else:
        assert captured.out == f"{src}\n"


def test_ls_root(cloud_test_catalog, cloud_type, capsys):
    src = cloud_test_catalog.src_uri
    root = src[: src.find("://") + 3]
    src_name = src.replace(root, "", 1)
    ls([root], catalog=cloud_test_catalog.catalog)
    captured = capsys.readouterr()
    if cloud_type == "file":
        assert not captured.out
    else:
        buckets = captured.out.split("\n")
        assert src_name in buckets


def ls_sources_output(src, cloud_type):
    if cloud_type == "file":
        root_uri = FileClient.root_path().as_uri()
        prefix = src[len(root_uri) :]
        return f"""\
{prefix}:
cats/
description
dogs/

{prefix}/dogs:
dog1
dog2
dog3

{prefix}/dogs/others:
dog4
    """

    return """\
cats/
dogs/
description

dogs/others:
dog4

dogs:
dog1
dog2
dog3
    """


def test_ls_sources(cloud_test_catalog, cloud_type, capsys):
    src = cloud_test_catalog.src_uri
    ls([src, f"{src}/dogs/*"], catalog=cloud_test_catalog.catalog)
    captured = capsys.readouterr()
    assert same_lines(captured.out, ls_sources_output(src, cloud_type))


def test_ls_sources_scheme_uppercased(cloud_test_catalog, cloud_type, capsys):
    src = uppercase_scheme(cloud_test_catalog.src_uri)
    ls([src, f"{src}/dogs/*"], catalog=cloud_test_catalog.catalog)
    captured = capsys.readouterr()
    assert same_lines(captured.out, ls_sources_output(src, cloud_type))


def test_ls_not_found(cloud_test_catalog):
    src = cloud_test_catalog.src_uri
    with pytest.raises(FileNotFoundError):
        ls([src, f"{src}/cats/bogus*"], catalog=cloud_test_catalog.catalog)


def test_ls_not_a_directory(cloud_test_catalog):
    src = cloud_test_catalog.src_uri
    with pytest.raises(FileNotFoundError):
        ls([src, f"{src}/description/"], catalog=cloud_test_catalog.catalog)


def ls_glob_output(src, cloud_type):
    if cloud_type == "file":
        root_uri = FileClient.root_path().as_uri()
        prefix = src[len(root_uri) :]
        return f"""\
{prefix}/dogs/others:
dog4

{prefix}/dogs:
dog1
dog2
dog3
    """

    return """\
dogs/others:
dog4

dogs:
dog1
dog2
dog3
    """


def test_ls_glob_sub(cloud_test_catalog, cloud_type, capsys):
    src = cloud_test_catalog.src_uri
    ls([f"{src}/dogs/*"], catalog=cloud_test_catalog.catalog)
    captured = capsys.readouterr()
    assert same_lines(captured.out, ls_glob_output(src, cloud_type))


def get_partial_indexed_paths(metastore):
    p = metastore._partials
    return [
        r[0] for r in metastore.db.execute(select(p.c.path_str).order_by(p.c.path_str))
    ]


def test_ls_partial_indexing(cloud_test_catalog, cloud_type, capsys):
    metastore = cloud_test_catalog.catalog.metastore
    src = cloud_test_catalog.src_uri
    if cloud_type == "file":
        src_metastore = metastore.clone(FileClient.root_path().as_uri())
        root_uri = FileClient.root_path().as_uri()
        prefix = src[len(root_uri) :] + "/"
    else:
        src_metastore = metastore.clone(src)
        prefix = ""

    ls([f"{src}/dogs/others/"], catalog=cloud_test_catalog.catalog)
    # These sleep calls are here to ensure that capsys can fully capture the output
    # and to avoid any flaky tests due to multithreading generating output out of order
    sleep(0.05)
    captured = capsys.readouterr()
    assert get_partial_indexed_paths(src_metastore) == [f"{prefix}dogs/others/"]
    assert "Listing" in captured.err
    assert captured.out == "dog4\n"

    ls([f"{src}/cats/"], catalog=cloud_test_catalog.catalog)
    sleep(0.05)
    captured = capsys.readouterr()
    assert get_partial_indexed_paths(src_metastore) == [
        f"{prefix}cats/",
        f"{prefix}dogs/others/",
    ]
    assert "Listing" in captured.err
    assert same_lines("cat1\ncat2\n", captured.out)

    ls([f"{src}/dogs/"], catalog=cloud_test_catalog.catalog)
    sleep(0.05)
    captured = capsys.readouterr()
    assert get_partial_indexed_paths(src_metastore) == [
        f"{prefix}cats/",
        f"{prefix}dogs/",
        f"{prefix}dogs/others/",
    ]
    assert "Listing" in captured.err
    assert same_lines("others/\ndog1\ndog2\ndog3\n", captured.out)

    ls([f"{src}/cats/"], catalog=cloud_test_catalog.catalog)
    sleep(0.05)
    captured = capsys.readouterr()
    assert get_partial_indexed_paths(src_metastore) == [
        f"{prefix}cats/",
        f"{prefix}dogs/",
        f"{prefix}dogs/others/",
    ]
    assert "Listing" not in captured.err
    assert same_lines("cat1\ncat2\n", captured.out)

    ls([f"{src}/"], catalog=cloud_test_catalog.catalog)
    sleep(0.05)
    captured = capsys.readouterr()
    assert get_partial_indexed_paths(src_metastore) == [
        f"{prefix}",
        f"{prefix}cats/",
        f"{prefix}dogs/",
        f"{prefix}dogs/others/",
    ]
    assert "Listing" in captured.err
    assert same_lines("cats/\ndogs/\ndescription\n", captured.out)


class MockResponse:
    def __init__(self, content, ok=True):
        self.content = content
        self.ok = ok


def mock_post(url, data=None, json=None, **kwargs):
    source = json["source"]
    path = re.sub(r"\w+://[^/]+/?", "", source).rstrip("/")
    data = [
        {
            **d,
            "path_str": d["path_str"].format(src=path),
            "path": d["path_str"].format(src=path).split("/"),
        }
        for d in REMOTE_DATA[path]
    ]
    return MockResponse(
        content=msgpack.packb({"data": data}, default=_pack_extended_types)
    )


def _pack_extended_types(obj):
    if isinstance(obj, datetime):
        if obj.tzinfo:
            data = (obj.timestamp(), int(obj.utcoffset().total_seconds()))
            return msgpack.ExtType(42, pack("!dl", *data))
        data = (obj.timestamp(),)
        return msgpack.ExtType(42, pack("!d", *data))
    raise TypeError(f"Unknown type: {type(obj)}")


ls_remote_sources_output = """\
{src}:
cats/
dogs/
description

{src}/dogs/others:
dog4

{src}/dogs:
dog1
dog2
dog3
"""


def test_ls_remote_sources(cloud_type, capsys, monkeypatch):
    src = f"{cloud_type}://bucket"
    token = "35NmrvSlsGVxTYIglxSsBIQHRrMpi6irSSYcAL0flijOytCHc"  # noqa: S105
    with monkeypatch.context() as m:
        m.setattr("requests.post", mock_post)
        ls(
            [src, f"{src}/dogs/others", f"{src}/dogs"],
            config={
                "type": "http",
                "url": "http://localhost:8111/api/dvcx",
                "username": "dvcx-team",
                "token": f"isat_{token}",
            },
        )
    captured = capsys.readouterr()
    assert captured.out == ls_remote_sources_output.format(src=src)


owner_id = "a13a3ff923430363b098ce9c769e450724e74e646332b08ca6b3ac4f96dae083"
REMOTE_DATA: dict[str, list[dict[str, Any]]] = {
    "": [
        {
            "id": 816,
            "dir_type": 1,
            "name": "cats",
            "etag": "",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2023, 1, 17, 21, 39, 0, 88564),
            "size": 0,
            "owner_name": "",
            "owner_id": "",
            "path_str": "{src}/cats",
            "path": [],
        },
        {
            "id": 825,
            "dir_type": 1,
            "name": "dogs",
            "etag": "",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2023, 1, 17, 21, 39, 0, 88567),
            "size": 0,
            "owner_name": "",
            "owner_id": "",
            "path_str": "{src}/dogs",
            "path": [],
        },
        {
            "id": None,
            "dir_type": 0,
            "name": "description",
            "etag": "20664550afa2654017377ceb266a1f82",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2022, 2, 10, 3, 39, 9),
            "size": 350496,
            "owner_name": "",
            "owner_id": owner_id,
            "path_str": "{src}/description",
            "path": [],
        },
    ],
    "dogs/others": [
        {
            "id": None,
            "dir_type": 0,
            "name": "dog4",
            "etag": "c4e42ce24d92bb5b4c4be9a99b237502",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2022, 6, 28, 22, 39, 1),
            "size": 32975,
            "owner_name": "",
            "owner_id": owner_id,
            "path_str": "{src}/dogs/others/dog4",
            "path": [],
        },
    ],
    "dogs": [
        {
            "id": None,
            "dir_type": 0,
            "name": "dog1",
            "etag": "44a632238558e0aa4c54bdb901bf9cff",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2022, 6, 28, 22, 39, 1),
            "size": 101,
            "owner_name": "",
            "owner_id": owner_id,
            "path_str": "{src}/dogs/dog1",
            "path": [],
        },
        {
            "id": None,
            "dir_type": 0,
            "name": "dog2",
            "etag": "76556c960239c50e5a8f8569daf85355",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2022, 6, 28, 22, 39, 1),
            "size": 29759,
            "owner_name": "",
            "owner_id": owner_id,
            "path_str": "{src}/dogs/dog2",
            "path": [],
        },
        {
            "id": None,
            "dir_type": 0,
            "name": "dog3",
            "etag": "b1c99fedcf77bf5fa62984e93db1955c",
            "version": "",
            "is_latest": True,
            "last_modified": datetime(2022, 6, 28, 22, 39, 1),
            "size": 102,
            "owner_name": "",
            "owner_id": owner_id,
            "path_str": "{src}/dogs/dog3",
            "path": [],
        },
    ],
}
